import collections.abc
import importlib
import typing

from decimal import Decimal
from functools import singledispatchmethod
from os import path
from typing import Optional, Sequence, Mapping, Union
from urllib.parse import urlparse

from . import types
from . import parser
from .errors import ProgrammingError
from .parser import ast

from .query_compile import (
    EvalAggregator,
    EvalAnd,
    EvalAll,
    EvalAny,
    EvalCoalesce,
    EvalColumn,
    EvalConstant,
    EvalCreateTable,
    EvalGetItem,
    EvalGetter,
    EvalInsert,
    EvalOr,
    EvalPivot,
    EvalQuery,
    EvalConstantSubquery1D,
    EvalRow,
    EvalTarget,
    FUNCTIONS,
    OPERATORS,
    SubqueryTable,
)

# Load functions and types definitions.
from . import query_env  # noqa: F401


# A global constant which sets whether we support inferred/implicit group-by
# semantics.
SUPPORT_IMPLICIT_GROUPBY = True


class CompilationError(ProgrammingError):
    def __init__(self, message, node=None):
        super().__init__(message)
        self.parseinfo = node.parseinfo if node is not None else None


class Compiler:
    def __init__(self, context):
        self.context = context
        self.stack = [context.tables.get(None)]

    @property
    def table(self):
        return self.stack[-1]

    @table.setter
    def table(self, value):
        self.stack[-1] = value

    def compile(self, query, parameters=None):
        """Compile an AST into an executable statement."""
        self.parameters = parameters

        placeholders = [node for node in query.walk() if isinstance(node, ast.Placeholder)]
        if placeholders:
            names = {placeholder.name for placeholder in placeholders}
            if all(names):
                if not isinstance(parameters, Mapping):
                    raise TypeError('query parameters should be a mapping when using named placeholders')
                if names - parameters.keys():
                    missing = ', '.join(sorted(names - parameters.keys()))
                    raise ProgrammingError(f'query parameter missing: {missing}')
            elif not any(names):
                if not isinstance(parameters, Sequence):
                    raise TypeError('query parameters should be a sequence when using positional placeholders')
                if len(placeholders) != len(parameters):
                    raise ProgrammingError(
                        f'the query has {len(placeholders)} placeholders but {len(parameters)} parameters were passed')
                for i, placeholder in enumerate(sorted(placeholders, key=lambda node: node.parseinfo.pos)):
                    placeholder.name = i
            else:
                raise ProgrammingError('positional and named parameters cannot be mixed')

        return self._compile(query)

    @singledispatchmethod
    def _compile(self, node: Optional[ast.Node]):
        if node is None:
            return None
        raise NotImplementedError

    @_compile.register
    def _select(self, node: ast.Select):
        self.stack.append(self.table)

        # Compile the FROM clause.
        c_from_expr = self._compile_from(node.from_clause)

        # Compile the targets.
        c_targets = self._compile_targets(node.targets)

        # Bind the WHERE expression to the execution environment.
        c_where = self._compile(node.where_clause)

        # Check that the FROM clause does not contain aggregates. This
        # should never trigger if the compilation environment does not
        # contain any aggregate.
        if c_where is not None and is_aggregate(c_where):
            raise CompilationError('aggregates are not allowed in WHERE clause')

        # Combine FROM and WHERE clauses
        if c_from_expr is not None:
            c_where = c_from_expr if c_where is None else EvalAnd([c_from_expr, c_where])

        # Process the GROUP-BY clause.
        new_targets, group_indexes, having_index = self._compile_group_by(node.group_by, c_targets)
        c_targets.extend(new_targets)

        # Process the ORDER-BY clause.
        new_targets, order_spec = self._compile_order_by(node.order_by, c_targets)
        c_targets.extend(new_targets)

        # If this is an aggregate query (it groups, see list of indexes), check that
        # the set of non-aggregates match exactly the group indexes. This should
        # always be the case at this point, because we have added all the necessary
        # targets to the list of group-by expressions and should have resolved all
        # the indexes.
        if group_indexes is not None:
            non_aggregate_indexes = {index for index, c_target in enumerate(c_targets)
                                     if not c_target.is_aggregate}
            if non_aggregate_indexes != set(group_indexes):
                missing_names = ['"{}"'.format(c_targets[index].name)
                                 for index in non_aggregate_indexes - set(group_indexes)]
                raise CompilationError(
                    'all non-aggregates must be covered by GROUP-BY clause in aggregate query: '
                    'the following targets are missing: {}'.format(','.join(missing_names)))

        query = EvalQuery(self.table,
                          c_targets,
                          c_where,
                          group_indexes,
                          having_index,
                          order_spec,
                          node.limit,
                          node.distinct)

        pivots = self._compile_pivot_by(node.pivot_by, c_targets, group_indexes)
        if pivots:
            return EvalPivot(query, pivots)

        self.stack.pop()
        return query

    def _compile_from(self, node):
        if node is None:
            return None

        # Subquery.
        if isinstance(node, ast.Select):
            self.table = SubqueryTable(self._compile(node))
            return None

        # Table reference.
        if isinstance(node, ast.Table):
            self.table = self.context.tables.get(node.name)
            if self.table is None:
                raise CompilationError(f'table "{node.name}" does not exist', node)
            return None

        # FROM expression.
        if isinstance(node, ast.From):
            # Check if the FROM expression is a column name belongin to the current table.
            if isinstance(node.expression, ast.Column):
                column = self.table.columns.get(node.expression.name)
                if column is None:
                    # When it is not, threat it as a table name.
                    table = self.context.tables.get(node.expression.name)
                    if table is not None:
                        self.table = table
                        return None

            c_expression = self._compile(node.expression)

            # Check that the FROM clause does not contain aggregates.
            if c_expression is not None and is_aggregate(c_expression):
                raise CompilationError('aggregates are not allowed in FROM clause')

            if node.open and node.close and node.open > node.close:
                raise CompilationError('CLOSE date must follow OPEN date')

            # Apply OPEN, CLOSE, and CLEAR clauses.
            if node.open is not None or node.close is not None or node.clear is not None:
                self.table = self.table.evolve(open=node.open, close=node.close, clear=node.clear)

            return c_expression

        raise NotImplementedError

    def _compile_targets(self, targets):
        """Compile the targets and check for their validity. Process wildcard.

        Args:
          targets: A list of target expressions from the parser.
        Returns:
          A list of compiled target expressions with resolved names.
        """
        # Bind the targets expressions to the execution context.
        if isinstance(targets, ast.Asterisk):
            # Insert the full list of available columns.
            targets = [ast.Target(ast.Column(name), None)
                       for name in self.table.wildcard_columns]

        # Compile targets.
        c_targets = []
        for target in targets:
            c_expr = self._compile(target.expression)
            name = get_target_name(target)
            c_targets.append(EvalTarget(c_expr, name, is_aggregate(c_expr)))

            columns, aggregates = get_columns_and_aggregates(c_expr)

            # Check for mixed aggregates and non-aggregates.
            if columns and aggregates:
                raise CompilationError('mixed aggregates and non-aggregates are not allowed')

            # Check for aggregates of aggregates.
            for aggregate in aggregates:
                for child in aggregate.childnodes():
                    if is_aggregate(child):
                        raise CompilationError('aggregates of aggregates are not allowed')

        return c_targets

    def _compile_order_by(self, order_by, c_targets):
        """Process an order-by clause.

        Args:
          order_by: A OrderBy instance as provided by the parser.
          c_targets: A list of compiled target expressions.
        Returns:
          A tuple of
           new_targets: A list of new compiled target nodes.
           order_spec: A list of (integer indexes, sort order) tuples.
        """
        if not order_by:
            return [], None

        # Compile order-by expressions and resolve them to their targets if
        # possible. A ORDER-BY column may be one of the following:
        #
        # * A reference to a target by name.
        # * A reference to a target by index (starting at one).
        # * A new expression, aggregate or not.
        #
        # References by name are converted to indexes. New expressions are
        # inserted into the list of targets as invisible targets.

        new_targets = c_targets[:]
        c_target_expressions = [c_target.c_expr for c_target in c_targets]
        targets_name_map = {target.name: idx for idx, target in enumerate(c_targets) if target.name is not None}
        # Only targets appearing in the SELECT targets list can be
        # referenced by index. These are guaranteed to have a valid name.
        n_targets = len(targets_name_map)

        order_spec = []

        for spec in order_by:
            column = spec.column
            descending = spec.ordering
            index = None

            # Process target references by index.
            if isinstance(column, int):
                index = column - 1
                if not 0 <= index < n_targets:
                    raise CompilationError(f'invalid ORDER-BY column index {column}')

            else:
                # Process target references by name. These will be parsed as
                # simple Column expressions. If they refer to a target name, we
                # resolve them.
                if isinstance(column, ast.Column):
                    name = column.name
                    index = targets_name_map.get(name, None)

                # Otherwise we compile the expression and add it to the list of
                # targets to evaluate and index into that new target.
                if index is None:
                    c_expr = self._compile(column)

                    # Attempt to reconcile the expression with one of the existing
                    # target expressions.
                    try:
                        index = c_target_expressions.index(c_expr)
                    except ValueError:
                        # Add the new target. 'None' for the target name implies it
                        # should be invisible, not to be rendered.
                        index = len(new_targets)
                        new_targets.append(EvalTarget(c_expr, None, is_aggregate(c_expr)))
                        c_target_expressions.append(c_expr)

            assert index is not None, "Internal error, could not index order-by reference."
            order_spec.append((index, descending))

        return new_targets[len(c_targets):], order_spec

    def _compile_pivot_by(self, pivot_by, targets, group_indexes):
        """Compiles a PIVOT BY clause.

        Resolve and validate columns references in the PIVOT BY clause.
        The PIVOT BY clause accepts two name od index references to
        columns in the SELECT targets list. The second columns should be a
        GROUP BY column so that the values of the pivot column are unique.

        """
        if pivot_by is None:
            return None

        indexes = []
        names = {target.name: index for index, target in enumerate(targets)}

        for column in pivot_by.columns:

            # Process target references by index.
            if isinstance(column, int):
                index = column - 1
                if not 0 <= index < len(targets):
                    raise CompilationError(f'invalid PIVOT BY column index {column}')
                indexes.append(index)
                continue

            # Process target references by name.
            if isinstance(column, ast.Column):
                index = names.get(column.name, None)
                if index is None:
                    raise CompilationError(f'PIVOT BY column {column!r} is not in the targets list')
                indexes.append(index)
                continue

            # Not reached.
            raise RuntimeError

        # Sanity checks.
        if indexes[0] == indexes[1]:
            raise CompilationError('the two PIVOT BY columns cannot be the same column')
        if indexes[1] not in group_indexes:
            raise CompilationError('the second PIVOT BY column must be a GROUP BY column')

        return indexes

    def _compile_group_by(self, group_by, c_targets):
        """Process a group-by clause.

        Args:
          group_by: A GroupBy instance as provided by the parser.
          c_targets: A list of compiled target expressions.
        Returns:
          A tuple of
           new_targets: A list of new compiled target nodes.
           group_indexes: If the query is an aggregate query, a list of integer
             indexes to be used for processing grouping. Note that this list may be
             empty (in the case of targets with only aggregates). On the other hand,
             if this is not an aggregated query, this is set to None. So do
             distinguish the empty list vs. None.
        """
        new_targets = c_targets[:]
        c_target_expressions = [c_target.c_expr for c_target in c_targets]

        group_indexes = []
        having_index = None

        if group_by:
            assert group_by.columns, "Internal error with GROUP-BY parsing"

            # Compile group-by expressions and resolve them to their targets if
            # possible. A GROUP-BY column may be one of the following:
            #
            # * A reference to a target by name.
            # * A reference to a target by index (starting at one).
            # * A new, non-aggregate expression.
            #
            # References by name are converted to indexes. New expressions are
            # inserted into the list of targets as invisible targets.
            targets_name_map = {target.name: index for index, target in enumerate(c_targets)}
            for column in group_by.columns:
                index = None

                # Process target references by index.
                if isinstance(column, int):
                    index = column - 1
                    if not 0 <= index < len(c_targets):
                        raise CompilationError(f'invalid GROUP-BY column index {column}')

                else:
                    # Process target references by name. These will be parsed as
                    # simple Column expressions. If they refer to a target name, we
                    # resolve them.
                    if isinstance(column, ast.Column):
                        name = column.name
                        index = targets_name_map.get(name, None)

                    # Otherwise we compile the expression and add it to the list of
                    # targets to evaluate and index into that new target.
                    if index is None:
                        c_expr = self._compile(column)

                        # Check if the new expression is an aggregate.
                        aggregate = is_aggregate(c_expr)
                        if aggregate:
                            raise CompilationError(f'GROUP-BY expressions may not be aggregates: "{column}"')

                        # Attempt to reconcile the expression with one of the existing
                        # target expressions.
                        try:
                            index = c_target_expressions.index(c_expr)
                        except ValueError:
                            # Add the new target. 'None' for the target name implies it
                            # should be invisible, not to be rendered.
                            index = len(new_targets)
                            new_targets.append(EvalTarget(c_expr, None, aggregate))
                            c_target_expressions.append(c_expr)

                assert index is not None, "Internal error, could not index group-by reference."
                group_indexes.append(index)

                # Check that the group-by column references a non-aggregate.
                c_expr = new_targets[index].c_expr
                if is_aggregate(c_expr):
                    raise CompilationError(f'GROUP-BY expressions may not reference aggregates: "{column}"')

                # Check that the group-by column has a supported hashable type.
                if not issubclass(c_expr.dtype, collections.abc.Hashable):
                    raise CompilationError(f'GROUP-BY a non-hashable type is not supported: "{column}"')

            # Compile HAVING clause.
            if group_by.having is not None:
                c_expr = self._compile(group_by.having)
                if not is_aggregate(c_expr):
                    raise CompilationError('the HAVING clause must be an aggregate expression')
                having_index = len(new_targets)
                new_targets.append(EvalTarget(c_expr, None, True))
                c_target_expressions.append(c_expr)

        else:
            # If it does not have a GROUP-BY clause...
            aggregate_bools = [c_target.is_aggregate for c_target in c_targets]
            if any(aggregate_bools):
                # If the query is an aggregate query, check that all the targets are
                # aggregates.
                if all(aggregate_bools):
                    # FIXME: shold we really be checking for the empty
                    # list or is checking for a false value enough?
                    assert group_indexes == []
                elif SUPPORT_IMPLICIT_GROUPBY:
                    # If some of the targets aren't aggregates, automatically infer
                    # that they are to be implicit group by targets. This makes for
                    # a much more convenient syntax for our lightweight SQL, where
                    # grouping is optional.
                    group_indexes = [
                        index for index, c_target in enumerate(c_targets)
                        if not c_target.is_aggregate]
                else:
                    raise CompilationError('aggregate query without a GROUP-BY should have only aggregates')
            else:
                # This is not an aggregate query; don't set group_indexes to
                # anything useful, we won't need it.
                group_indexes = None

        return new_targets[len(c_targets):], group_indexes, having_index

    @_compile.register
    def _column(self, node: ast.Column):
        column = self.table.columns.get(node.name)
        if column is not None:
            return column
        raise CompilationError(f'column "{node.name}" not found in table "{self.table.name}"', node)

    @_compile.register
    def _or(self, node: ast.Or):
        return EvalOr([self._compile(arg) for arg in node.args])

    @_compile.register
    def _and(self, node: ast.And):
        return EvalAnd([self._compile(arg) for arg in node.args])

    _OPERATORS = {
        '<': ast.Less,
        '<=': ast.LessEq,
        '>': ast.Greater,
        '>=': ast.GreaterEq,
        '=': ast.Equal,
        '!=': ast.NotEqual,
        '~': ast.Match,
        '!~': ast.NotMatch,
        '?~': ast.Matches,
    }

    # dispatching on an Union is supported only starting with Python 3.11
    @_compile.register(ast.All)
    @_compile.register(ast.Any)
    def _all(self, node):
        right = self._compile(node.right)

        if isinstance(right, EvalQuery):
            if len(right.columns) != 1:
                raise CompilationError('subquery has too many columns', node.right)
            right = EvalConstantSubquery1D(right)

        right_dtype = typing.get_origin(right.dtype) or right.dtype
        if right_dtype not in {list, set}:
            raise CompilationError(f'not a list or set but {right_dtype}', node.right)
        args = typing.get_args(right.dtype)
        if args:
            assert len(args) == 1
            right_element_dtype = args[0]
        else:
            right_element_dtype = object

        left = self._compile(node.left)

        # lookup operator implementaton and check typing
        op = self._OPERATORS[node.op]
        for func in OPERATORS[op]:
            if func.__intypes__ == [right_element_dtype, left.dtype]:
                break
        else:
            raise CompilationError(
                f'operator "{op.__name__.lower()}('
                f'{left.dtype.__name__}, {right_element_dtype.__name__})" not supported', node)

        # need to instantiate the operaotr implementation to get to the underlying function
        operator = func(None, None).operator

        cls = EvalAll if type(node) is ast.All else EvalAny
        return cls(operator, left, right)

    @_compile.register
    def _function(self, node: ast.Function):
        operands = [self._compile(operand) for operand in node.operands]

        # ``row(*)`` is parsed like a function call but does something special
        if node.fname == 'row' and len(operands) == 1 and operands[0].dtype == types.Asterisk:
            return EvalRow()

        # ``coalesce()`` is parsed like a function call but it does
        # not really fit our model for function evaluation, therefore
        # it gets special threatment here.
        if node.fname == 'coalesce':
            for operand in operands:
                if operand.dtype != operands[0].dtype:
                    dtypes = ', '.join(operand.dtype.__name__ for operand in operands)
                    raise CompilationError(f'coalesce() function arguments must have uniform type, found: {dtypes}', node)
            return EvalCoalesce(operands)

        function = types.function_lookup(FUNCTIONS, node.fname, operands)
        if function is None:
            sig = '{}({})'.format(node.fname, ', '.join(f'{operand.dtype.__name__.lower()}' for operand in operands))
            raise CompilationError(f'no function matches "{sig}" name and argument types', node)

        # Replace ``meta(key)`` with ``meta[key]``.
        if node.fname == 'meta':
            key = node.operands[0]
            node = ast.Function('getitem', [ast.Column('meta', parseinfo=node.parseinfo), key])
            return self._compile(node)

        # Replace ``entry_meta(key)`` with ``entry.meta[key]``.
        if node.fname == 'entry_meta':
            key = node.operands[0]
            node = ast.Function('getitem', [ast.Attribute(ast.Column('entry', parseinfo=node.parseinfo), 'meta'), key])
            return self._compile(node)

        # Replace ``any_meta(key)`` with ``getitem(meta, key, entry.meta[key])``.
        if node.fname == 'any_meta':
            key = node.operands[0]
            node = ast.Function('getitem', [ast.Column('meta', parseinfo=node.parseinfo), key, ast.Function('getitem', [
                ast.Attribute(ast.Column('entry', parseinfo=node.parseinfo), 'meta'), key])])
            return self._compile(node)

        # Replace ``has_account(regexp)`` with ``('(?i)' + regexp) ~? any (accounts)``.
        if node.fname == 'has_account':
            node = ast.Any(ast.Add(ast.Constant('(?i)'), node.operands[0]), '?~', ast.Column('accounts'))
            return self._compile(node)

        function = function(self.context, operands)
        # Constants folding.
        if all(isinstance(operand, EvalConstant) for operand in operands) and function.pure:
            return EvalConstant(function(None), function.dtype)
        return function

    @_compile.register
    def _subscript(self, node: ast.Subscript):
        operand = self._compile(node.operand)
        if issubclass(operand.dtype, dict):
            return EvalGetItem(operand, node.key)
        raise CompilationError('column type is not subscriptable', node)

    @_compile.register
    def _attribute(self, node: ast.Attribute):
        operand = self._compile(node.operand)
        dtype = types.ALIASES.get(operand.dtype, operand.dtype)
        if issubclass(dtype, types.Structure):
            getter = dtype.columns.get(node.name)
            if getter is None:
                raise CompilationError(f'structured type has no attribute "{node.name}"', node)
            return EvalGetter(operand, getter, getter.dtype)
        raise CompilationError('column type is not structured', node)

    @_compile.register
    def _unaryop(self, node: ast.UnaryOp):
        operand = self._compile(node.operand)
        function = types.function_lookup(OPERATORS, type(node), [operand])
        if function is None:
            raise CompilationError(
                f'operator "{type(node).__name__.lower()}({types.name(operand.dtype)})" not supported', node)
        function = function(operand)
        # Constants folding.
        if isinstance(operand, EvalConstant):
            return EvalConstant(function(None), function.dtype)
        return function

    @_compile.register
    def _between(self, node: ast.Between):
        operand = self._compile(node.operand)
        lower = self._compile(node.lower)
        upper = self._compile(node.upper)
        intypes = [operand.dtype, lower.dtype, upper.dtype]
        for candidate in OPERATORS[type(node)]:
            if candidate.__intypes__ == intypes:
                func = candidate(operand, lower, upper)
                return func
        raise CompilationError(
            f'operator "{types.name(operand.dtype)} BETWEEN {types.name(lower.dtype)} '
            f'AND {types.name(upper.dtype)}" not supported', node)

    @_compile.register(ast.In)
    @_compile.register(ast.NotIn)
    def _inop(self, node: Union[ast.In, ast.NotIn]):
        left = self._compile(node.left)
        right = self._compile(node.right)

        if isinstance(right, EvalQuery):
            if len(right.columns) != 1:
                raise CompilationError('subquery has too many columns', node.right)
            right = EvalConstantSubquery1D(right)

        op = OPERATORS[type(node)][0]
        return op(left, right)

    @_compile.register
    def _binaryop(self, node: ast.BinaryOp):
        left = self._compile(node.left)
        right = self._compile(node.right)

        candidates = OPERATORS[type(node)]
        while True:
            intypes = [left.dtype, right.dtype]
            for op in candidates:
                if op.__intypes__ == intypes:
                    function = op(left, right)
                    # Constants folding.
                    if isinstance(left, EvalConstant) and isinstance(right, EvalConstant):
                        return EvalConstant(function(None), function.dtype)
                    return function

            # Implement type inference when one of the operands is not strongly typed.
            if left.dtype is object and right.dtype is not object:
                target = right.dtype
                if target is int:
                    # The Beancount parser does not emit int typed
                    # values, thus casting to int is only going to
                    # loose information. Promote to decimal.
                    target = Decimal
                name = types.MAP.get(target)
                if name is None:
                    break
                left = types.function_lookup(FUNCTIONS, name, [left])(self.context, [left])
                continue
            if right.dtype is object and left.dtype is not object:
                target = left.dtype
                if target is int:
                    # The Beancount parser does not emit int typed
                    # values, thus casting to int is only going to
                    # loose information. Promote to decimal.
                    target = Decimal
                name = types.MAP.get(target)
                if name is None:
                    break
                right = types.function_lookup(FUNCTIONS, name, [right])(self.context, [right])
                continue

            # Failure.
            break

        raise CompilationError(
            f'operator "{type(node).__name__.lower()}('
            f'{types.name(left.dtype)}, {types.name(right.dtype)})" not supported', node)

    @_compile.register
    def _constant(self, node: ast.Constant):
        # For backward compatibility, the parser allows strings to be
        # delimited by single or double quotes. This creates ambiguity between
        # quoted identifiers and string. Strings delimited by double quotes
        # are treated as column names when they resolve to an existing column
        # in the current table.
        if isinstance(node.value, str) and node.text and node.text[0] == '"':
            if node.value in self.table.columns:
                return self._column(ast.Column(node.value))
        return EvalConstant(node.value)

    @_compile.register
    def _placeholder(self, node: ast.Placeholder):
        return EvalConstant(self.parameters[node.name])

    @_compile.register
    def _asterisk(self, node: ast.Asterisk):
        return EvalConstant(None, dtype=types.Asterisk)

    @_compile.register
    def _balances(self, node: ast.Balances):
        return self._compile(transform_balances(node))

    @_compile.register
    def _journal(self, node: ast.Journal):
        return self._compile(transform_journal(node))

    @_compile.register
    def _print(self, node: ast.Print):
        self.table = self.context.tables.get('entries')
        expr = self._compile_from(node.from_clause)
        targets = [EvalTarget(EvalRow(), 'ROW(*)', False)]
        return EvalQuery(self.table, targets, expr, None, None, None, None, False)

    @_compile.register
    def _create_table(self, node: ast.CreateTable):
        query = None
        columns = None
        if node.columns is not None:
            columns = []
            for cname, ctype in node.columns:
                datatype = types.parse(ctype)
                if datatype is None:
                    raise CompilationError(f'unrecognized type "{ctype}"', node)
                columns.append((cname, datatype))
        if node.query is not None:
            query = self._compile(node.query)
            columns = [(t.name, t.c_expr.dtype) for t in query.c_targets if t.name is not None]
        parts = urlparse(node.using)
        scheme = parts.scheme or path.splitext(parts.path)[1][1:] or 'memory'
        impl = importlib.import_module(f'beanquery.sources.{scheme}').create
        return EvalCreateTable(self.context, node.name, columns, node.using, query, impl)

    @_compile.register
    def _insert(self, node: ast.Insert):
        table = self.context.tables.get(node.table.name)
        if table is None:
            raise CompilationError(f'table "{node.table.name}" does not exist', node.table)
        impl = getattr(table, 'insert', None)
        if impl is None:
            raise CompilationError(f'table "{node.table.name}" does not support insertion', node.table)
        if len(node.values) != len(node.columns):
            raise CompilationError(
                f'column names and values mismatch: '
                f'expected {len(node.columns)} but {len(node.values)} values were supplied', node)
        values = [EvalConstant(None)] * len(table.columns)
        columns = {name: i for i, name in enumerate(table.columns.keys())}
        for column, value in zip(node.columns, node.values):
            index = columns.get(column.name)
            if index is None:
                raise CompilationError(f'column "{column.name}" not found in table "{node.table.name}"', column)
            expr = self._compile(value)
            if not expr.dtype == table.columns.get(column.name).dtype:
                raise CompilationError(f'expression has wrong type for column "{column.name}"', value)
            values[index] = expr
        return EvalInsert(table, values)


def transform_journal(journal):
    """Translate a Journal entry into an uncompiled Select statement.

    Args:
      journal: An instance of a Journal object.
    Returns:
      An instance of an uncompiled Select object.
    """
    cooked_select = parser.parse("""

        SELECT
           date,
           flag,
           MAXWIDTH(payee, 48),
           MAXWIDTH(narration, 80),
           account,
           {summary_func}(position),
           {summary_func}(balance)
        {where}

    """.format(where=('WHERE account ~ "{}"'.format(journal.account)
                      if journal.account
                      else ''),
               summary_func=journal.summary_func or ''))

    return ast.Select(cooked_select.targets,
                      journal.from_clause,
                      cooked_select.where_clause,
                      None, None, None, None, None)


def transform_balances(balances):
    """Translate a Balances entry into an uncompiled Select statement.

    Args:
      balances: An instance of a Balance object.
    Returns:
      An instance of an uncompiled Select object.
    """
    ## FIXME: Change the aggregation rules to allow GROUP-BY not to include the
    ## non-aggregate ORDER-BY columns, so we could just GROUP-BY accounts here
    ## instead of having to include the sort-key. I think it should be fine if
    ## the first or last sort-order value gets used, because it would simplify
    ## the input statement.

    cooked_select = parser.parse("""

      SELECT account, SUM({}(position))
      GROUP BY account, ACCOUNT_SORTKEY(account)
      ORDER BY ACCOUNT_SORTKEY(account)

    """.format(balances.summary_func or ""))

    return ast.Select(cooked_select.targets,
                      balances.from_clause,
                      balances.where_clause,
                      cooked_select.group_by,
                      cooked_select.order_by,
                      None, None, None)


def get_target_name(target):
    """Compute the target name.

    This uses the same algorithm used by SQLite. If the target has an
    AS clause assigning it a name, that will be the name used. If the
    target refers directly to a column, then the target name is the
    column name. Otherwise use the expression text.

    """
    if target.name is not None:
        return target.name
    if isinstance(target.expression, ast.Column):
        return target.expression.name
    return target.expression.text.strip()


def get_columns_and_aggregates(node):
    """Find the columns and aggregate nodes below this tree.

    All nodes under aggregate nodes are ignored.

    Args:
      node: An instance of EvalNode.
    Returns:
      A pair of (columns, aggregates), both of which are lists of EvalNode instances.
        columns: The list of all columns accessed not under an aggregate node.
        aggregates: The list of all aggregate nodes.
    """
    columns = []
    aggregates = []
    _get_columns_and_aggregates(node, columns, aggregates)
    return columns, aggregates


def _get_columns_and_aggregates(node, columns, aggregates):
    """Walk down a tree of nodes and fetch the column accessors and aggregates.

    This function ignores all nodes under aggregate nodes.

    Args:
      node: An instance of EvalNode.
      columns: An accumulator for columns found so far.
      aggregate: An accumulator for aggregate notes found so far.
    """
    if isinstance(node, EvalAggregator):
        aggregates.append(node)
    elif isinstance(node, EvalColumn):
        columns.append(node)
    else:
        for child in node.childnodes():
            _get_columns_and_aggregates(child, columns, aggregates)


def is_aggregate(node):
    """Return true if the node is an aggregate.

    Args:
      node: An instance of EvalNode.
    Returns:
      A boolean.
    """
    # Note: We could be a tiny bit more efficient here, but it doesn't matter
    # much. Performance of the query compilation matters very little overall.
    _, aggregates = get_columns_and_aggregates(node)
    return bool(aggregates)


def compile(context, statement, parameters=None):
    return Compiler(context).compile(statement, parameters)



project = 'beanquery'
copyright = '2014-2022, beanquery Contributors'
author = 'beanquery Contributors'
version = '0.1'
language = 'en'
html_theme = 'furo'
html_title = f'{project} {version}'
html_logo = 'logo.svg'
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.githubpages',
]
extlinks = {
    'issue': ('https://github.com/beancount/beanquery/issues/%s', '#'),
    'pull': ('https://github.com/beancount/beanquery/pull/%s', '#'),
}
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'beancount': ('https://beancount.github.io/docs/', None),
}
napoleon_google_docstring = True
napoleon_use_param = False
autodoc_typehints = 'none'
autodoc_member_order = 'bysource'



__copyright__ = "Copyright (C) 2014-2017  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import unittest
import textwrap

from decimal import Decimal
from dateutil.relativedelta import relativedelta

from beancount.core.amount import from_string as A
from beancount.core.number import D
from beancount.core import inventory
from beancount.core.inventory import from_string as I
from beancount.parser import cmptest
from beancount.utils.test_utils import docfile
from beancount import loader

import beanquery

from beanquery import CompilationError
from beanquery import query_compile as qc
from beanquery import query_execute as qx
from beanquery import tables
from beanquery import compiler


class QueryBase(cmptest.TestCase):
    INPUT = ""
    maxDiff = 8192

    def setUp(self):
        entries, errors, options = loader.load_string(textwrap.dedent(self.INPUT))
        self.ctx = beanquery.connect('beancount:', entries=entries, errors=errors, options=options)

    def compile(self, query):
        return self.ctx.compile(self.ctx.parse(query))

    def check_query(self, input_string, query, expected_types, expected_rows):
        entries, errors, options = loader.load_string(textwrap.dedent(input_string))
        ctx = beanquery.connect('beancount:', entries=entries, errors=errors, options=options)
        curs = ctx.execute(query)
        self.assertEqual(tuple(expected_types), curs.description)
        result_rows = curs.fetchall()
        self.assertEqual(expected_rows, result_rows)

    def assertResult(self, query, result, datatype=None):
        curs = self.ctx.execute(query)
        self.assertEqual([c.datatype for c in curs.description], [datatype or type(result)])
        self.assertEqual(curs.fetchall(), [(result, )])

    def assertError(self, query):
        with self.assertRaises(CompilationError):
            self.ctx.execute(query)


class CommonInputBase(unittest.TestCase):
    INPUT = textwrap.dedent("""

    2010-01-01 open Assets:Bank:Checking
    2010-01-01 open Assets:ForeignBank:Checking
    2010-01-01 open Assets:Bank:Savings

    2010-01-01 open Expenses:Restaurant

    2010-01-01 * "Dinner with Cero"
      Assets:Bank:Checking       100.00 USD
      Expenses:Restaurant       -100.00 USD

    2011-01-01 * "Dinner with Uno"
      Assets:Bank:Checking       101.00 USD
      Expenses:Restaurant       -101.00 USD

    2012-02-02 * "Dinner with Dos"
      Assets:Bank:Checking       102.00 USD
      Expenses:Restaurant       -102.00 USD

    2013-03-03 * "Dinner with Tres"
      Assets:Bank:Checking       103.00 USD
      Expenses:Restaurant       -103.00 USD

    2013-10-10 * "International Transfer"
      Assets:Bank:Checking         -50.00 USD
      Assets:ForeignBank:Checking  -60.00 CAD @ 1.20 USD

    2014-04-04 * "Dinner with Quatro"
      Assets:Bank:Checking       104.00 USD
      Expenses:Restaurant       -104.00 USD

    """)


class TestFundamentals(QueryBase):
    INPUT = textwrap.dedent("""
          2022-04-05 commodity TEST
            rate: 42
          2022-04-05 open Assets:Tests
          2022-04-05 * "Test"
            Assets:Tests  1.000 TEST
              int: 1
              decimal: 1.2
              bool: TRUE
              str: "str"
              str3: "3"
              str4: "4.0"
              date: 2022-04-05
              null: NULL
        """)

    def test_type_casting(self):
        # bool
        self.assertResult("SELECT bool(TRUE)", True)
        self.assertResult("SELECT bool(1)", True)
        self.assertResult("SELECT bool(1.1)", True)
        self.assertResult("SELECT bool('foo')", True)
        self.assertResult("SELECT bool(NULL)", None, bool)
        self.assertResult("SELECT bool(2022-04-05)", True)
        self.assertResult("SELECT bool(meta('int'))", True)
        self.assertResult("SELECT bool(meta('decimal'))", True)
        self.assertResult("SELECT bool(meta('bool'))", True)
        self.assertResult("SELECT bool(meta('str'))", True)
        self.assertResult("SELECT bool(meta('date'))", True)
        self.assertResult("SELECT bool(meta('null'))", None, bool)
        self.assertResult("SELECT bool(meta('missing'))", None, bool)

        # int
        self.assertResult("SELECT int(TRUE)", 1)
        self.assertResult("SELECT int(1)", 1)
        self.assertResult("SELECT int(1.2)", 1)
        self.assertResult("SELECT int('1')", 1)
        self.assertResult("SELECT int('foo')", None, int)
        self.assertResult("SELECT int(NULL)", None, int)
        self.assertError ("SELECT int(2022-04-05)")
        self.assertResult("SELECT int(meta('int'))", 1)
        self.assertResult("SELECT int(meta('decimal'))", 1)
        self.assertResult("SELECT int(meta('bool'))", 1)
        self.assertResult("SELECT int(meta('str'))", None, int)
        self.assertResult("SELECT int(meta('str3'))", 3)
        self.assertResult("SELECT int(meta('str4'))", None, int)
        self.assertResult("SELECT int(meta('date'))", None, int)
        self.assertResult("SELECT int(meta('null'))", None, int)
        self.assertResult("SELECT int(meta('missing'))", None, int)

        # decimal
        self.assertResult("SELECT decimal(TRUE)", Decimal(1))
        self.assertResult("SELECT decimal(1)", Decimal(1))
        self.assertResult("SELECT decimal(1.2)", Decimal('1.2'))
        self.assertResult("SELECT decimal('1.2')", Decimal('1.2'))
        self.assertResult("SELECT decimal('foo')", None, Decimal)
        self.assertResult("SELECT decimal(NULL)", None, Decimal)
        self.assertError ("SELECT decimal(2022-04-05)")
        self.assertResult("SELECT decimal(meta('int'))", Decimal(1))
        self.assertResult("SELECT decimal(meta('decimal'))", Decimal('1.2'))
        self.assertResult("SELECT decimal(meta('bool'))", Decimal(1))
        self.assertResult("SELECT decimal(meta('str'))", None, Decimal)
        self.assertResult("SELECT decimal(meta('str3'))", Decimal(3))
        self.assertResult("SELECT decimal(meta('str4'))", Decimal('4.0'))
        self.assertResult("SELECT decimal(meta('date'))", None, Decimal)
        self.assertResult("SELECT decimal(meta('null'))", None, Decimal)
        self.assertResult("SELECT decimal(meta('missing'))", None, Decimal)

        # str
        self.assertResult("SELECT str(TRUE)", 'TRUE')
        self.assertResult("SELECT str(FALSE)", 'FALSE')
        self.assertResult("SELECT str(1)", '1')
        self.assertResult("SELECT str(1.1)", '1.1')
        self.assertResult("SELECT str('foo')", 'foo')
        self.assertResult("SELECT str(NULL)", None, str)
        self.assertResult("SELECT str(2022-04-05)", '2022-04-05')
        self.assertResult("SELECT str(meta('int'))", '1')
        self.assertResult("SELECT str(meta('decimal'))", '1.2')
        self.assertResult("SELECT str(meta('bool'))", 'TRUE')
        self.assertResult("SELECT str(meta('str'))", 'str')
        self.assertResult("SELECT str(meta('date'))", '2022-04-05')
        self.assertResult("SELECT str(meta('null'))", None, str)
        self.assertResult("SELECT str(meta('missing'))", None, str)

        # date
        self.assertError ("SELECT date(TRUE)")
        self.assertError ("SELECT date(1)")
        self.assertError ("SELECT date(1.2)")
        self.assertResult("SELECT date('1.2')", None, datetime.date)
        self.assertResult("SELECT date('foo')", None, datetime.date)
        self.assertResult("SELECT date('2022-04-05')", datetime.date(2022, 4, 5))
        self.assertResult("SELECT date(NULL)", None, datetime.date)
        self.assertResult("SELECT date(2022-04-05)", datetime.date(2022, 4, 5))
        self.assertResult("SELECT date(2022, 4, 5)", datetime.date(2022, 4, 5))
        self.assertResult("SELECT date(meta('int'))", None, datetime.date)
        self.assertResult("SELECT date(meta('decimal'))", None, datetime.date)
        self.assertResult("SELECT date(meta('bool'))", None, datetime.date)
        self.assertResult("SELECT date(meta('str'))", None, datetime.date)
        self.assertResult("SELECT date(meta('date'))", datetime.date(2022, 4, 5))
        self.assertResult("SELECT date(meta('null'))", None, datetime.date)
        self.assertResult("SELECT date(meta('missing'))", None, datetime.date)

    def test_operators(self):
        # add
        self.assertResult("SELECT 1 + 1", 2)
        self.assertResult("SELECT 1.0 + 1", Decimal(2))
        self.assertResult("SELECT 1.0 + 2.00", Decimal(3))
        self.assertError ("SELECT 1970-01-01 + 2022-04-01")
        self.assertResult("SELECT 2022-04-01 + 1", datetime.date(2022, 4, 2))
        self.assertResult("SELECT 1 + 2022-04-01", datetime.date(2022, 4, 2))
        self.assertResult("SELECT 'ab' + 'c'", 'abc')

        # sub
        self.assertResult("SELECT 1 - 1", 0)
        self.assertResult("SELECT 1.0 - 1", Decimal(0))
        self.assertResult("SELECT 1.0 - 2.00", Decimal(-1))
        self.assertResult("SELECT 2022-04-01 - 1", datetime.date(2022, 3, 31))
        self.assertResult("SELECT 2022-04-01 - 2022-03-31", 1)
        self.assertError ("SELECT 1 - 2022-04-01")

        # mul
        self.assertResult("SELECT 2 * 2", 4)
        self.assertResult("SELECT 2.0 * 2", Decimal(4))
        self.assertResult("SELECT 2 * 2.0", Decimal(4))
        self.assertResult("SELECT 2.0 * 2.0", Decimal(4))

        # div
        self.assertResult("SELECT 4 / 2", Decimal(2))
        self.assertResult("SELECT 4.0 / 2", Decimal(2))
        self.assertResult("SELECT 4 / 2.0", Decimal(2))
        self.assertResult("SELECT 4.0 / 2.0", Decimal(2))
        self.assertResult("SELECT 4 / 0", None, Decimal)
        self.assertResult("SELECT 4.0 / 0", None, Decimal)
        self.assertResult("SELECT 4 / 0.0", None, Decimal)

        # mod
        self.assertResult("SELECT 3 % 2", 1)
        self.assertResult("SELECT 3.0 % 2", Decimal(1))
        self.assertResult("SELECT 3 % 2.0", Decimal(1))
        self.assertResult("SELECT 3.0 % 2.0", Decimal(1))

        # match
        self.assertResult("SELECT 'foobarbaz' ~ 'bar'", True)
        self.assertResult("SELECT 'foobarbaz' ~ 'quz'", False)

        # mot match
        self.assertResult("SELECT 'foobarbaz' !~ 'bar'", False)
        self.assertResult("SELECT 'foobarbaz' !~ 'quz'", True)

        # matches
        self.assertResult("SELECT '[0-9]+' ?~ '123'", True)
        self.assertResult("SELECT '[0-9]+' ?~ 'ABC'", False)

        # and
        self.assertResult("SELECT 1 and FALSE", False)
        self.assertResult("SELECT 'something' and FALSE", False)
        self.assertResult("SELECT 1.0 and FALSE", False)
        self.assertResult("SELECT TRUE and meta('missing')", None, bool)
        self.assertResult("SELECT TRUE and not meta('missing')", True)

        # or
        self.assertResult("SELECT FALSE or 1", True)
        self.assertResult("SELECT FALSE or 'something'", True)
        self.assertResult("SELECT FALSE or 1.0", True)
        self.assertResult("SELECT TRUE or meta('missing')", True)
        self.assertResult("SELECT FALSE or not meta('missing')", True)

        # not
        self.assertResult("SELECT not TRUE", False)
        self.assertResult("SELECT not meta('missing')", True)

        # is null
        self.assertResult("SELECT date IS NULL", False)
        self.assertResult("SELECT meta('missing') IS NULL", True)
        self.assertResult("SELECT meta('int') IS NULL", False)

        # is not null
        self.assertResult("SELECT date IS NOT NULL", True)
        self.assertResult("SELECT meta('missing') IS NOT NULL", False)
        self.assertResult("SELECT meta('int') IS NOT NULL", True)

        # in
        self.assertResult("SELECT 'tag' IN tags", False)
        self.assertResult("SELECT 3 IN (2, 3, 4)", True)
        self.assertResult("SELECT 'x' IN ('a', 'b', 'c')", False)
        self.assertResult("SELECT 3 NOT IN (2, 3, 4)", False)
        self.assertResult("SELECT 1 NOT IN (2, 3, 4)", True)

        # between
        self.assertResult("SELECT 2 BETWEEN 1 AND 3", True)
        self.assertResult("SELECT 1 BETWEEN 2 AND 3", False)
        self.assertResult("SELECT 2.0 BETWEEN 1.0 AND 3.0", True)
        self.assertResult("SELECT 1.0 BETWEEN 2 AND 3.0", False)
        self.assertResult("SELECT 2023-08-02 BETWEEN 2023-08-01 AND 2023-08-03", True)
        self.assertResult("SELECT 'b' BETWEEN 'a' AND 'c'", True)
        self.assertError ("SELECT 2023-08-02 BETWEEN 'a' AND 2")

    def test_operators_type_inference(self):
        self.assertResult("SELECT 1 + meta('int')", Decimal(2))
        self.assertResult("SELECT 1 + meta('str3')", Decimal(4))
        self.assertResult("SELECT meta('int') > 0", True)

    def test_functions(self):
        # round
        self.assertResult("SELECT round(1.2)", Decimal(1))
        self.assertResult("SELECT round(1.234, 2)", Decimal('1.23'))
        self.assertResult("SELECT round(12)", 12)
        self.assertResult("SELECT round(12, -1)", 10)

        # neg
        self.assertResult("SELECT neg(1.0)", Decimal('-1.0'))

        # abs
        self.assertResult("SELECT abs(-1.0)", Decimal('1.0'))

        # safediv
        self.assertResult("SELECT safediv(1.0, 2.0)", Decimal('0.5'))
        self.assertResult("SELECT safediv(1.0, 0.0)", Decimal('0'))
        self.assertResult("SELECT safediv(0.0, 0.0)", Decimal('0'))

        # repr
        self.assertResult("SELECT repr(1.0)", 'Decimal(\'1.0\')')

        # lower
        self.assertResult('''SELECT lower('AaBbCcDdEe')''', 'aabbccddee')

        # upper
        self.assertResult('''SELECT upper('AaBbCcDdEe')''', 'AABBCCDDEE')

    def test_coalesce(self):
        # coalesce
        self.assertResult("SELECT COALESCE(str(meta('missing')), '!')", "!")
        self.assertError ("SELECT COALESCE(meta('missing'), '!')")

    def test_count(self):
        # count number of postings
        self.assertResult("SELECT COUNT(*)", 1)
        self.assertResult("SELECT COUNT(1)", 1)
        self.assertResult("SELECT COUNT(NULL)", 0)
        self.assertResult("SELECT COUNT(meta('missing'))", 0)

    def test_getitem(self):
        self.assertResult("SELECT meta['lineno']", 6, object)
        self.assertResult("SELECT 1 + meta['lineno']", Decimal('7'))
        self.assertResult("SELECT meta['filename']", '<string>', object)
        self.assertResult("SELECT meta['missing']", None, object)
        self.assertError ("SELECT position['foo']")

    def test_getatrr(self):
        self.assertResult("SELECT entry.flag", '*')
        self.assertResult("SELECT int(entry.meta['lineno'])", 5)
        self.assertError ("SELECT entry.foo")
        self.assertError ("SELECT position.foo")

    def test_parse_date(self):
        self.assertResult('''SELECT parse_date('2016/11/1')''', datetime.date(2016, 11, 1))
        self.assertResult('''SELECT parse_date('2016/11/1', '%Y/%d/%m')''', datetime.date(2016, 1, 11))

    def test_date_part(self):
        self.assertResult('''SELECT date_part('weekday', 2024-06-09)''', 6)
        self.assertResult('''SELECT date_part('dow', 2024-06-09)''', 6)
        self.assertResult('''SELECT date_part('isoweekday', 2024-06-09)''', 7)
        self.assertResult('''SELECT date_part('isodow', 2024-06-09)''', 7)
        self.assertResult('''SELECT date_part('week', 2024-06-09)''', 23)
        self.assertResult('''SELECT date_part('month', 2024-06-09)''', 6)
        self.assertResult('''SELECT date_part('quarter', 2024-06-09)''', 2)
        self.assertResult('''SELECT date_part('year', 2024-06-09)''', 2024)
        self.assertResult('''SELECT date_part('isoyear', 2024-06-09)''', 2024)
        self.assertResult('''SELECT date_part('decade', 2024-06-09)''', 202)
        self.assertResult('''SELECT date_part('century', 2024-06-09)''', 21)
        self.assertResult('''SELECT date_part('millennium', 2024-06-09)''', 3)
        self.assertResult('''SELECT date_part('epoch', 2024-06-09)''', 1717891200)
        self.assertResult('''SELECT date_part('baz', 2024-06-09)''', None, int)

    def test_date_trunc(self):
        self.assertResult('''SELECT date_trunc('week', 2016-11-14)''', datetime.date(2016, 11, 14)) # monday
        self.assertResult('''SELECT date_trunc('week', 2016-11-15)''', datetime.date(2016, 11, 14)) # tuesday
        self.assertResult('''SELECT date_trunc('week', 2016-11-20)''', datetime.date(2016, 11, 14)) # sunday

        self.assertResult('''SELECT date_trunc('month', 2016-11-20)''', datetime.date(2016, 11, 1))

        self.assertResult('''SELECT date_trunc('quarter', 2016-09-30)''', datetime.date(2016, 7, 1))
        self.assertResult('''SELECT date_trunc('quarter', 2016-10-01)''', datetime.date(2016, 10, 1))
        self.assertResult('''SELECT date_trunc('quarter', 2016-11-20)''', datetime.date(2016, 10, 1))

        self.assertResult('''SELECT date_trunc('year', 2016-11-20)''', datetime.date(2016, 1, 1))

        self.assertResult('''SELECT date_trunc('decade', 2016-11-20)''', datetime.date(2010, 1, 1))
        self.assertResult('''SELECT date_trunc('decade', 2020-11-20)''', datetime.date(2020, 1, 1))
        self.assertResult('''SELECT date_trunc('decade', 2029-11-20)''', datetime.date(2020, 1, 1))

        self.assertResult('''SELECT date_trunc('century', 1999-11-20)''', datetime.date(1901, 1, 1))
        self.assertResult('''SELECT date_trunc('century', 2000-11-20)''', datetime.date(1901, 1, 1))
        self.assertResult('''SELECT date_trunc('century', 2001-11-20)''', datetime.date(2001, 1, 1))
        self.assertResult('''SELECT date_trunc('century', 2016-11-20)''', datetime.date(2001, 1, 1))

        self.assertResult('''SELECT date_trunc('millennium', 1991-11-20)''', datetime.date(1001, 1, 1))
        self.assertResult('''SELECT date_trunc('millennium', 2000-11-20)''', datetime.date(1001, 1, 1))
        self.assertResult('''SELECT date_trunc('millennium', 2001-11-20)''', datetime.date(2001, 1, 1))
        self.assertResult('''SELECT date_trunc('millennium', 2016-11-20)''', datetime.date(2001, 1, 1))
        self.assertResult('''SELECT date_trunc('millennium', 3456-11-20)''', datetime.date(3001, 1, 1))

        self.assertResult('''SELECT date_trunc('foo', 2024-05-24)''', None, datetime.date)

    def test_interval_ops(self):
        self.assertResult('''SELECT 2016-11-20 + interval("1 months")''', datetime.date(2016, 12, 20))
        self.assertResult('''SELECT 2016-11-01 + interval("1 months") - interval("1 days")''', datetime.date(2016, 11, 30))
        self.assertResult('''SELECT 2024-02-01 + interval("1 month") + interval("-1 day")''', datetime.date(2024, 2, 29))
        self.assertResult('''SELECT 2024-02-05 + interval("-1 days")''', datetime.date(2024, 2, 4))
        self.assertResult('''SELECT 2024-02-05 + interval("1 day") + interval("2 days")''', datetime.date(2024, 2, 8))
        self.assertResult('''SELECT 2024-05-24 + interval("1 year")''', datetime.date(2025, 5, 24))
        self.assertResult('''SELECT 2024-05-24 + interval("-2 years")''', datetime.date(2022, 5, 24))
        self.assertResult('''SELECT interval("1 baz")''', None, relativedelta)
        self.assertResult('''SELECT interval("A days")''', None, relativedelta)

    def test_date_bin(self):
        self.assertResult('''SELECT date_bin(interval('1 year'), 2024-11-10, 2024-06-01)''', datetime.date(2024, 6, 1))
        self.assertResult('''SELECT date_bin('1 year', 2024-11-10, 2024-06-01)''', datetime.date(2024, 6, 1))
        self.assertResult('''SELECT date_bin('1 year', 2024-11-10, 2025-06-01)''', datetime.date(2024, 6, 1))
        self.assertResult('''SELECT date_bin('1 month', 2024-11-10, 2024-06-03)''', datetime.date(2024, 11, 3))
        self.assertResult('''SELECT date_bin('3 days', 2024-11-10, 2024-11-02)''', datetime.date(2024, 11, 8))
        self.assertResult('''SELECT date_bin('3 days', 2024-11-10, 2024-11-14)''', datetime.date(2024, 11, 8))


class TestBeancountFunctions(QueryBase):
    INPUT = textwrap.dedent("""
          2024-06-23 commodity TEST
            answer: "42"
          2024-06-23 open Assets:Tests
            key: "value"
          2024-06-23 open Expenses:Tests
          2024-06-23 price USD 2.0 TEST
          2024-06-23 * "Test"
            Assets:Tests  10.000 TEST
            Expenses:Tests
          2024-06-24 close Expenses:Tests
        """)

    def test_root(self):
        self.assertResult('''SELECT root('Assets:Foo:Bar', 2) FROM #''', 'Assets:Foo')
        self.assertResult('''SELECT root('Assets:Foo:Bar', 1) FROM #''', 'Assets')
        self.assertResult('''SELECT root('Assets:Foo:Bar') FROM #''', 'Assets')

    def test_open_date(self):
        self.assertResult('''SELECT open_date('Expenses:Tests') FROM #''', datetime.date(2024, 6, 23))
        self.assertResult('''SELECT open_date('Expenses:Foo') FROM #''', None, datetime.date)

    def test_close_date(self):
        self.assertResult('''SELECT close_date('Expenses:Tests') FROM #''', datetime.date(2024, 6, 24))
        self.assertResult('''SELECT close_date('Expenses:Foo') FROM #''', None, datetime.date)
        self.assertResult('''SELECT close_date('Assets:Tests') FROM #''', None, datetime.date)

    def test_open_meta(self):
        self.assertResult('''SELECT open_meta('Assets:Tests') FROM #''', {'filename': '<string>', 'lineno': 4, 'key': 'value'})
        self.assertResult('''SELECT open_meta('Assets:Tests', 'key') FROM #''', 'value', object)
        self.assertResult('''SELECT open_meta('Expenses:Tests', 'baz') FROM #''', None, object)
        self.assertResult('''SELECT open_meta('Expenses:Foo', 'key') FROM #''', None, object)

    def test_commodity_meta(self):
        self.assertResult('''SELECT commodity_meta('TEST', 'answer') FROM #''', '42', object)
        self.assertResult('''SELECT commodity_meta('TEST') FROM #''', {'filename': '<string>', 'lineno': 2, 'answer': '42'})
        self.assertResult('''SELECT commodity_meta('X') FROM #''', None, dict)

    def test_account_sortkey(self):
        self.assertResult('''SELECT account_sortkey('Assets:Foo') LIMIT 1''', '0-Assets:Foo')

    def test_has_account(self):
        self.assertResult('''SELECT has_account('Assets:Tests') LIMIT 1''', True)
        self.assertResult('''SELECT has_account('Assets:Foo') LIMIT 1''', False)

    def test_convert(self):
        self.assertResult('''SELECT convert(position, 'USD') LIMIT 1''', A('5.0 USD'))
        self.assertResult('''SELECT convert(sum(position), 'USD')''', I('0.0 USD'))
        self.assertResult('''SELECT convert(only('TEST', sum(position)), 'USD')''', A('0.0 USD'))


class TestBeancountMetadataFunctions(QueryBase):
    INPUT = """
        2016-11-20 * "Test"
          name: "The Name"
          address: "1 Wrong Way"
          empty: "Not Empty"
          Assets:Banking          1 USD
            color: "Green"
            address: "1 Right Way"
            empty:
        """

    def test_meta(self):
        self.assertResult('''SELECT meta('address')''', '1 Right Way', object)
        self.assertResult('''SELECT meta('baz')''', None, object)
        self.assertResult('''SELECT meta('color')''', 'Green', object)
        self.assertResult('''SELECT meta('empty')''', None, object)
        self.assertResult('''SELECT meta('name')''', None, object)

    def test_entry_meta(self):
        self.assertResult('''SELECT entry_meta('address')''', '1 Wrong Way', object)
        self.assertResult('''SELECT entry_meta('baz')''', None, object)
        self.assertResult('''SELECT entry_meta('color')''', None, object)
        self.assertResult('''SELECT entry_meta('empty')''', 'Not Empty', object)
        self.assertResult('''SELECT entry_meta('name')''', 'The Name', object)

    def test_any_meta(self):
        self.assertResult('''SELECT any_meta('address')''', '1 Right Way', object)
        self.assertResult('''SELECT any_meta('baz')''', None, object)
        self.assertResult('''SELECT any_meta('color')''', 'Green', object)
        self.assertResult('''SELECT any_meta('empty')''', None, object)
        self.assertResult('''SELECT any_meta('name')''', 'The Name', object)


class TestFilterEntries(CommonInputBase, QueryBase):

    def compile(self, query):
        # use the ``entries`` table as default table for queries
        c = compiler.Compiler(self.ctx)
        c.table = self.ctx.tables.get('entries')
        return c.compile(self.ctx.parse(query))

    @staticmethod
    def filter_entries(query):
        entries = []
        expr = query.c_where
        for entry in query.table:
            if expr is None or expr(entry):
                entries.append(entry)
        return entries

    def test_filter_empty_from(self):
        # Check that no filter outputs the very same thing.
        filtered_entries = self.filter_entries(self.compile("""
          SELECT * ;
        """))
        self.assertEqualEntries(self.INPUT, filtered_entries)

    def test_filter_by_year(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM year(date) = 2012;
        """))
        self.assertEqualEntries("""

          2012-02-02 * "Dinner with Dos"
            Assets:Bank:Checking              102.00 USD
            Expenses:Restaurant              -102.00 USD

        """, filtered_entries)

    def test_filter_by_expr1(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type
          FROM NOT (type = 'transaction' AND
                    (year(date) = 2012 OR year(date) = 2013));
        """))
        self.assertEqualEntries("""

          2010-01-01 open Assets:Bank:Checking
          2010-01-01 open Assets:Bank:Savings
          2010-01-01 open Expenses:Restaurant
          2010-01-01 open Assets:ForeignBank:Checking

          2010-01-01 * "Dinner with Cero"
            Assets:Bank:Checking              100.00 USD
            Expenses:Restaurant              -100.00 USD

          2011-01-01 * "Dinner with Uno"
            Assets:Bank:Checking              101.00 USD
            Expenses:Restaurant              -101.00 USD

          2014-04-04 * "Dinner with Quatro"
            Assets:Bank:Checking              104.00 USD
            Expenses:Restaurant              -104.00 USD

        """, filtered_entries)

    def test_filter_by_expr2(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM date < 2012-06-01;
        """))
        self.assertEqualEntries("""

          2010-01-01 open Assets:Bank:Checking
          2010-01-01 open Assets:Bank:Savings
          2010-01-01 open Expenses:Restaurant
          2010-01-01 open Assets:ForeignBank:Checking

          2010-01-01 * "Dinner with Cero"
            Assets:Bank:Checking              100.00 USD
            Expenses:Restaurant              -100.00 USD

          2011-01-01 * "Dinner with Uno"
            Assets:Bank:Checking              101.00 USD
            Expenses:Restaurant              -101.00 USD

          2012-02-02 * "Dinner with Dos"
            Assets:Bank:Checking              102.00 USD
            Expenses:Restaurant              -102.00 USD

        """, filtered_entries)

    def test_filter_close_undated(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM CLOSE;
        """))

        self.assertEqualEntries(self.INPUT + textwrap.dedent("""

          2014-04-04 C "Conversion for (-50.00 USD, -60.00 CAD)"
            Equity:Conversions:Current  50.00 USD @ 0 NOTHING
            Equity:Conversions:Current  60.00 CAD @ 0 NOTHING

        """), filtered_entries)

    def test_filter_close_dated(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM CLOSE ON 2013-06-01;
        """))
        self.assertEqualEntries("""

          2010-01-01 open Assets:Bank:Checking
          2010-01-01 open Assets:ForeignBank:Checking
          2010-01-01 open Assets:Bank:Savings

          2010-01-01 open Expenses:Restaurant

          2010-01-01 * "Dinner with Cero"
            Assets:Bank:Checking       100.00 USD
            Expenses:Restaurant       -100.00 USD

          2011-01-01 * "Dinner with Uno"
            Assets:Bank:Checking       101.00 USD
            Expenses:Restaurant       -101.00 USD

          2012-02-02 * "Dinner with Dos"
            Assets:Bank:Checking       102.00 USD
            Expenses:Restaurant       -102.00 USD

          2013-03-03 * "Dinner with Tres"
            Assets:Bank:Checking       103.00 USD
            Expenses:Restaurant       -103.00 USD

        """, filtered_entries)

    def test_filter_open_dated(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM OPEN ON 2013-01-01;
        """))

        self.assertEqualEntries("""

          2010-01-01 open Assets:Bank:Checking
          2010-01-01 open Assets:Bank:Savings
          2010-01-01 open Expenses:Restaurant
          2010-01-01 open Assets:ForeignBank:Checking

          2012-12-31 S "Opening balance for 'Assets:Bank:Checking' (Summarization)"
            Assets:Bank:Checking          303.00 USD
            Equity:Opening-Balances      -303.00 USD

          2012-12-31 S "Opening balance for 'Equity:Earnings:Previous' (Summarization)"
            Equity:Earnings:Previous     -303.00 USD
            Equity:Opening-Balances       303.00 USD

          2013-03-03 * "Dinner with Tres"
            Assets:Bank:Checking          103.00 USD
            Expenses:Restaurant          -103.00 USD

          2013-10-10 * "International Transfer"
            Assets:Bank:Checking          -50.00 USD            ; -50.00 USD
            Assets:ForeignBank:Checking   -60.00 CAD @ 1.20 USD ; -72.00 USD

          2014-04-04 * "Dinner with Quatro"
            Assets:Bank:Checking          104.00 USD
            Expenses:Restaurant          -104.00 USD

        """, filtered_entries)

    def test_filter_clear(self):
        filtered_entries = self.filter_entries(self.compile("""
          SELECT date, type FROM CLEAR;
        """))

        self.assertEqualEntries(self.INPUT + textwrap.dedent("""

          2014-04-04 T "Transfer balance for 'Expenses:Restaurant' (Transfer balance)"
            Expenses:Restaurant                                 510.00 USD
            Equity:Earnings:Current                            -510.00 USD

        """), filtered_entries)


class TestAllocation(unittest.TestCase):

    def test_allocator(self):
        allocator = qx.Allocator()
        self.assertEqual(0, allocator.allocate())
        self.assertEqual(1, allocator.allocate())
        self.assertEqual(2, allocator.allocate())
        self.assertEqual([None, None, None], allocator.create_store())


class TestExecuteNonAggregatedQuery(QueryBase):

    INPUT = """

      2010-01-01 open Assets:Bank:Checking
      2010-01-01 open Expenses:Restaurant

      2010-02-23 * "Bla"
        Assets:Bank:Checking       100.00 USD
        Expenses:Restaurant       -100.00 USD

    """

    def test_non_aggregate__one(self):
        self.check_query(
            self.INPUT,
            """
            SELECT date;
            """,
            [('date', datetime.date)],
            [(datetime.date(2010, 2, 23),),
             (datetime.date(2010, 2, 23),)])

    def test_non_aggregate__many(self):
        self.check_query(
            self.INPUT,
            """
            SELECT date, flag, payee, narration;
            """,
            [
                ('date', datetime.date),
                ('flag', str),
                ('payee', str),
                ('narration', str),
                ],
            [
                (datetime.date(2010, 2, 23), '*', None, 'Bla'),
                (datetime.date(2010, 2, 23), '*', None, 'Bla'),
                ])

    def test_non_aggregated_order_by_visible(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account, length(account) ORDER BY 2;
            """,
            [
                ('account', str),
                ('length(account)', int),
                ],
            [
                ('Expenses:Restaurant', 19),
                ('Assets:Bank:Checking', 20),
                ])

    def test_non_aggregated_order_by_invisible(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account ORDER BY length(account);
            """,
            [
                ('account', str),
                ],
            [
                ('Expenses:Restaurant',),
                ('Assets:Bank:Checking',),
                ])

    def test_non_aggregated_order_by_none_date(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account ORDER BY cost_date;
            """,
            [
                ('account', str),
                ],
            [
                ('Assets:Bank:Checking',),
                ('Expenses:Restaurant',),
                ])

    def test_non_aggregated_order_by_none_str(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account ORDER BY posting_flag;
            """,
            [
                ('account', str),
                ],
            [
                ('Assets:Bank:Checking',),
                ('Expenses:Restaurant',),
                ])


class TestExecuteAggregatedQuery(QueryBase):

    INPUT = """

      2010-01-01 open Assets:Bank:Checking
      2010-01-01 open Expenses:Restaurant

      2010-02-23 * "Bla"
        Assets:Bank:Checking       100.00 USD
        Expenses:Restaurant       -100.00 USD

    """

    def test_aggregated_group_by_all_implicit(self):
        # There is no group-by, but all columns are aggregations.
        self.check_query(
            self.INPUT,
            """
            SELECT first(account), last(account);
            """,
            [
                ('first(account)', str),
                ('last(account)', str),
                ],
            [
                ('Assets:Bank:Checking', 'Expenses:Restaurant'),
                ])

    def test_aggregated_group_by_all_explicit(self):
        # All columns ('account', 'len') are subject of a group-by.
        self.check_query(
            self.INPUT,
            """
            SELECT account, length(account) as len
            GROUP BY account, len;
            """,
            [
                ('account', str),
                ('len', int),
                ],
            [
                ('Assets:Bank:Checking', 20),
                ('Expenses:Restaurant', 19),
                ])

        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT account, length(account) as len
            GROUP BY account, len;
            """,
            [
                ('account', str),
                ('len', int),
                ],
            [
                ('Assets:Bank:Checking', 20),
                ('Expenses:Restaurant', 19),
                ('Liabilities:Credit-Card', 23),
                ])

    def test_aggregated_group_by_visible(self):
        # GROUP-BY: 'account' is visible.
        self.check_query(
            self.INPUT,
            """
            SELECT account, sum(position) as amount
            GROUP BY account;
            """,
            [
                ('account', str),
                ('amount', inventory.Inventory),
                ],
            [
                ('Assets:Bank:Checking', inventory.from_string('100.00 USD')),
                ('Expenses:Restaurant', inventory.from_string('-100.00 USD')),
                ])

    def test_aggregated_group_by_invisible(self):
        # GROUP-BY: 'account' is invisible.
        self.check_query(
            self.INPUT,
            """
            SELECT count(position)
            GROUP BY account;
            """,
            [
                ('count(position)', int),
                ],
            [
                (1,),
                (1,),
                ])

    def test_aggregated_group_by_visible_order_by_non_aggregate_visible(self):
        # GROUP-BY: 'account' is visible.
        # ORDER-BY: 'account' is a non-aggregate and visible.
        self.check_query(
            self.INPUT,
            """
            SELECT account, sum(position) as amount
            GROUP BY account
            ORDER BY account;
            """,
            [
                ('account', str),
                ('amount', inventory.Inventory),
                ],
            [
                ('Assets:Bank:Checking', inventory.from_string('100.00 USD')),
                ('Expenses:Restaurant', inventory.from_string('-100.00 USD')),
                ])

    def test_aggregated_group_by_visible_order_by_non_aggregate_invisible(self):
        # GROUP-BY: 'account' and 'length(account)' are visible.
        # ORDER-BY: 'length(account)' is a non-aggregate and invisible.
        self.check_query(
            self.INPUT,
            """
            SELECT account, sum(position) as amount
            GROUP BY account, length(account)
            ORDER BY length(account);
            """,
            [
                ('account', str),
                ('amount', inventory.Inventory),
                ],
            [
                ('Expenses:Restaurant', inventory.from_string('-100.00 USD')),
                ('Assets:Bank:Checking', inventory.from_string('100.00 USD')),
                ])

    def test_aggregated_group_by_visible_order_by_aggregate_visible(self):
        # GROUP-BY: 'account' is visible.
        # ORDER-BY: 'sum(account)' is an aggregate and visible.
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT account, count(account) as num, sum(number) as sum
            GROUP BY account
            ORDER BY sum(number);
            """,
            [
                ('account', str),
                ('num', int),
                ('sum', Decimal),
                ],
            [
                ('Liabilities:Credit-Card', 1, D('-2.00')),
                ('Assets:Bank:Checking', 1, D('-1.00')),
                ('Expenses:Restaurant', 2, D('3.00')),
                ])

    def test_aggregated_group_by_visible_order_by_aggregate_invisible(self):
        # GROUP-BY: 'account' is visible.
        # ORDER-BY: 'sum(number)' is an aggregate and invisible.
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT account, count(account) as num
            GROUP BY account
            ORDER BY sum(number);
            """,
            [
                ('account', str),
                ('num', int),
                ],
            [
                ('Liabilities:Credit-Card', 1),
                ('Assets:Bank:Checking', 1),
                ('Expenses:Restaurant', 2),
                ])

    def test_aggregated_group_by_invisible_order_by_non_aggregate_visible(self):
        # GROUP-BY: 'account' is invisible.
        # ORDER-BY: 'len(account)' is a non-aggregate and visible.
        self.check_query(
            self.INPUT,
            """
            SELECT length(account) as len, sum(position) as amount
            GROUP BY account, len
            ORDER BY len;
            """,
            [
                ('len', int),
                ('amount', inventory.Inventory),
                ],
            [
                (19, inventory.from_string('-100.00 USD'),),
                (20, inventory.from_string('100.00 USD'),),
                ])

    def test_aggregated_group_by_invisible_order_by_non_aggregate_invis(self):
        # GROUP-BY: 'account' is invisible.
        # ORDER-BY: 'sum(number)' is an aggregate and invisible.
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT count(account) as num
            GROUP BY account
            ORDER BY sum(number);
            """,
            [
                ('num', int),
                ],
            [
                (1,),
                (1,),
                (2,),
                ])

    def test_aggregated_group_by_invisible_order_by_aggregate_visible(self):
        # GROUP-BY: 'account' is invisible.
        # ORDER-BY: 'sum(account)' is an aggregate and visible.
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT count(account) as num, sum(number) as sum
            GROUP BY account
            ORDER BY sum(number);
            """,
            [
                ('num', int),
                ('sum', Decimal),
                ],
            [
                (1, D('-2.00')),
                (1, D('-1.00')),
                (2, D('3.00')),
                ])

    def test_aggregated_group_by_invisible_order_by_aggregate_invisible(self):
        # GROUP-BY: 'account' is invisible.
        # ORDER-BY: 'sum(number)' is an aggregate and invisible.
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Restaurant         1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Restaurant         2.00 USD
            """,
            """
            SELECT count(account) as num
            GROUP BY account
            ORDER BY sum(number);
            """,
            [
                ('num', int),
                ],
            [
                (1,),
                (1,),
                (2,),
                ])

    def test_aggregated_group_by_with_having(self):
        self.check_query(
            """
            2010-02-21 * "First"
              Assets:Bank:Checking       -1.00 USD
              Expenses:Foo                1.00 USD

            2010-02-23 * "Second"
              Liabilities:Credit-Card    -2.00 USD
              Expenses:Bar                2.00 USD
            """,
            """
            SELECT account, sum(number)
            GROUP BY account
            HAVING sum(number) > 0.0
            ORDER BY account
            """,
            [
                ('account', str),
                ('sum(number)', Decimal),
            ],
            [
                ('Expenses:Bar', D(2.0)),
                ('Expenses:Foo', D(1.0)),
            ])


class TestExecuteOptions(QueryBase):

    INPUT = """

      2010-02-23 *
        Assets:AssetA       5.00 USD
        Assets:AssetD       2.00 USD
        Assets:AssetB       4.00 USD
        Assets:AssetC       3.00 USD
        Assets:AssetE       1.00 USD
        Equity:Rest       -15.00 USD

    """

    def test_order_by_asc_implicit(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account, number ORDER BY number;
            """,
            [
                ('account', str),
                ('number', Decimal),
                ],
            [
                ('Equity:Rest', D('-15.00')),
                ('Assets:AssetE', D('1.00')),
                ('Assets:AssetD', D('2.00')),
                ('Assets:AssetC', D('3.00')),
                ('Assets:AssetB', D('4.00')),
                ('Assets:AssetA', D('5.00')),
                ])

    def test_order_by_asc_explicit(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account, number ORDER BY number ASC;
            """,
            [
                ('account', str),
                ('number', Decimal),
                ],
            [
                ('Equity:Rest', D('-15.00')),
                ('Assets:AssetE', D('1.00')),
                ('Assets:AssetD', D('2.00')),
                ('Assets:AssetC', D('3.00')),
                ('Assets:AssetB', D('4.00')),
                ('Assets:AssetA', D('5.00')),
                ])

    def test_order_by_desc(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account, number ORDER BY number DESC;
            """,
            [
                ('account', str),
                ('number', Decimal),
                ],
            [
                ('Assets:AssetA', D('5.00')),
                ('Assets:AssetB', D('4.00')),
                ('Assets:AssetC', D('3.00')),
                ('Assets:AssetD', D('2.00')),
                ('Assets:AssetE', D('1.00')),
                ('Equity:Rest', D('-15.00')),
                ])

    def test_distinct(self):
        self.check_query(
            """
              2010-02-23 *
                Assets:AssetA       5.00 USD
                Assets:AssetA       2.00 USD
                Assets:AssetA       4.00 USD
                Equity:Rest
            """,
            """
            SELECT account ;
            """,
            [
                ('account', str),
                ],
            [
                ('Assets:AssetA',),
                ('Assets:AssetA',),
                ('Assets:AssetA',),
                ('Equity:Rest',),
                ])

        self.check_query(
            """
              2010-02-23 *
                Assets:AssetA       5.00 USD
                Assets:AssetA       2.00 USD
                Assets:AssetA       4.00 USD
                Equity:Rest        -5.00 USD
                Equity:Rest        -2.00 USD
                Equity:Rest        -4.00 USD
            """,
            """
            SELECT DISTINCT account ;
            """,
            [
                ('account', str),
                ],
            [
                ('Assets:AssetA',),
                ('Equity:Rest',),
                ])

    def test_limit(self):
        self.check_query(
            self.INPUT,
            """
            SELECT account, number ORDER BY number LIMIT 3;
            """,
            [
                ('account', str),
                ('number', Decimal),
                ],
            [
                ('Equity:Rest', D('-15.00')),
                ('Assets:AssetE', D('1.00')),
                ('Assets:AssetD', D('2.00')),
                ])


class TestOrderBy(QueryBase):
    data = """

    2022-03-28 * "Test"
      Assets:Tests  1.00 USD
        aa: 1
        bb: 1
      Assets:Tests  2.00 USD
        aa: 1
        bb: 2
      Assets:Tests  3.00 USD
        aa: 2
      Assets:Tests  4.00 USD
        aa: 2
        bb: 1
      Expenses:Tests

    """

    def test_order_by_asc_asc(self):
        self.check_query(self.data,
            """SELECT account, meta('aa') AS a, meta('bb') AS b ORDER BY 2, 3""",
            [
                ('account', str), ('a', object), ('b', object)
            ],
            [
                ('Expenses:Tests', None, None),
                ('Assets:Tests', 1, 1),
                ('Assets:Tests', 1, 2),
                ('Assets:Tests', 2, None),
                ('Assets:Tests', 2, 1),
            ])

    def test_order_by_asc_desc(self):
        self.check_query(self.data,
            """SELECT account, meta('aa') AS a, meta('bb') AS b ORDER BY 2, 3 DESC""",
            [
                ('account', str), ('a', object), ('b', object)
            ],
            [
                ('Expenses:Tests', None, None),
                ('Assets:Tests', 1, 2),
                ('Assets:Tests', 1, 1),
                ('Assets:Tests', 2, 1),
                ('Assets:Tests', 2, None),
            ])

    def test_order_by_desc_asc(self):
        self.check_query(self.data,
            """SELECT account, meta('aa') AS a, meta('bb') AS b ORDER BY 2 DESC, 3""",
            [
                ('account', str), ('a', object), ('b', object)
            ],
            [
                ('Assets:Tests', 2, None),
                ('Assets:Tests', 2, 1),
                ('Assets:Tests', 1, 1),
                ('Assets:Tests', 1, 2),
                ('Expenses:Tests', None, None),
            ])

    def test_order_by_desc_desc(self):
        self.check_query(self.data,
            """SELECT account, meta('aa') AS a, meta('bb') AS b ORDER BY 2 DESC, 3 DESC""",
            [
                ('account', str), ('a', object), ('b', object)
            ],
            [
                ('Assets:Tests', 2, 1),
                ('Assets:Tests', 2, None),
                ('Assets:Tests', 1, 2),
                ('Assets:Tests', 1, 1),
                ('Expenses:Tests', None, None),
            ])


class TestArithmeticFunctions(QueryBase):

    # You need some transactions in order to eval a simple arithmetic op.
    # This also properly sets the data type to Decimal.
    # In v2, revise this so that this works like a regular DB and support integers.

    def test_add(self):
        self.check_query(
            """
              2010-02-23 *
                Assets:Something       5.00 USD
            """,
            """
              SELECT number + 3 as result;
            """,
            [('result', Decimal)],
            [(D("8"),)])

    def test_sub(self):
        self.check_query(
            """
              2010-02-23 *
                Assets:Something       5.00 USD
            """,
            """
              SELECT number - 3 as result;
            """,
            [('result', Decimal)],
            [(D("2"),)])

    def test_mul(self):
        self.check_query(
            """
              2010-02-23 *
                Assets:Something       5.00 USD
            """,
            """
              SELECT number * 1.2 as result;
            """,
            [('result', Decimal)],
            [(D("6"),)])

    def test_div(self):
        self.check_query(
            """
              2010-02-23 *
                Assets:Something       5.00 USD
            """,
            """
              SELECT number / 2 as result;
            """,
            [('result', Decimal)],
            [(D("2.50"),)])

        self.check_query(
            """
              2010-02-23 *
                Assets:Something       5.00 USD
            """,
            """
              SELECT number / 0 as result;
            """,
            [('result', Decimal)],
            [(None,)])


class TestExecutePivot(QueryBase):

    def setUp(self):
        super().setUp()
        entries, errors, options = loader.load_string(self.data, dedent=True)
        self.assertFalse(errors)
        self.ctx = beanquery.connect('beancount:', entries=entries, errors=errors, options=options)

    def execute(self, query):
        curs = self.ctx.execute(query)
        rows = curs.fetchall()
        columns = [(c.name, c.datatype) for c in curs.description]
        return columns, rows

    data = """
      2012-01-01 open Assets:Cash
      2012-01-01 open Expenses:Aaa
      2012-01-01 open Expenses:Bbb

      2012-01-01 * "Test"
        Expenses:Bbb  1.00 USD
        Assets:Cash

      2012-01-01 * "Test"
        Expenses:Aaa  2.00 USD
        Assets:Cash

      2012-02-02 * "Test"
        Expenses:Aaa  3.00 USD
        Assets:Cash

      2013-01-01 * "Test"
        Expenses:Aaa  4.00 USD
        Assets:Cash

      2014-02-02 * "Test"
        Expenses:Bbb  5.00 USD
        Assets:Cash

      2014-03-03 * "Test"
        Expenses:Aaa  6.00 USD
        Assets:Cash

      2013-01-01 * "Test"
        Expenses:Bbb  7.00 USD
        Assets:Cash

      2015-04-04 * "Test"
        Expenses:Aaa  8.00 USD
        Assets:Cash
    """

    def test_pivot_one_column(self):
        self.assertEqual(self.execute("""
            SELECT
              account,
              year(date) AS year,
              sum(cost(position)) AS balance
            WHERE
              account ~ 'Expenses'
            GROUP BY 1, 2
            PIVOT BY 1, 2"""), (
            [
                ('account/year', str),
                ('2012', inventory.Inventory),
                ('2013', inventory.Inventory),
                ('2014', inventory.Inventory),
                ('2015', inventory.Inventory),
            ],
            [
                ('Expenses:Aaa', I('5.00 USD'), I('4.00 USD'), I('6.00 USD'), I('8.00 USD')),
                ('Expenses:Bbb', I('1.00 USD'), I('7.00 USD'), I('5.00 USD'), None),
            ]))

    def test_pivot_one_column_by_name(self):
        self.assertEqual(self.execute("""
            SELECT
              account,
              year(date) AS year,
              sum(cost(position)) AS balance
            WHERE
              account ~ 'Expenses'
            GROUP BY 1, 2
            PIVOT BY account, year"""), (
            [
                ('account/year', str),
                ('2012', inventory.Inventory),
                ('2013', inventory.Inventory),
                ('2014', inventory.Inventory),
                ('2015', inventory.Inventory),
            ],
            [
                ('Expenses:Aaa', I('5.00 USD'), I('4.00 USD'), I('6.00 USD'), I('8.00 USD')),
                ('Expenses:Bbb', I('1.00 USD'), I('7.00 USD'), I('5.00 USD'), None),
            ]))

    def test_pivot_two_column(self):
        self.assertEqual(self.execute("""
            SELECT
              account,
              year(date) AS year,
              sum(cost(position)) AS balance,
              last(date) AS updated
            WHERE
              account ~ 'Expenses' AND
              year >= 2014
            GROUP BY 1, 2
            PIVOT BY 1, 2"""), (
            [
                ('account/year', str),
                ('2014/balance', inventory.Inventory),
                ('2014/updated', datetime.date),
                ('2015/balance', inventory.Inventory),
                ('2015/updated', datetime.date),
            ],
            [
                ('Expenses:Aaa', I('6.00 USD'), datetime.date(2014, 3, 3), I('8.00 USD'), datetime.date(2015, 4, 4)),
                ('Expenses:Bbb', I('5.00 USD'), datetime.date(2014, 2, 2), None, None),
            ]))


class TestExecuteSubquery(QueryBase):

    def execute(self, query):
        curs = self.ctx.execute(query)
        rows = curs.fetchall()
        columns = [(c.name, c.datatype) for c in curs.description]
        return columns, rows

    def test_subquery(self):
        self.assertEqual(
            self.execute("""SELECT a + 2 AS b FROM (SELECT 3 AS a FROM #)"""),
            ([('b', int)], [(5, )]))


class SimpleColumn(qc.EvalColumn):
    def __init__(self, name, func, dtype):
        super().__init__(dtype)
        self.name = name
        self.func = func

    def __call__(self, row):
        return self.func(row)


class SimpleTable(tables.Table):
    columns = {
        'column': SimpleColumn('column', lambda row: row, int)
    }

    def __init__(self, name, nrows):
        self.name = name
        self.nrows = nrows

    def __iter__(self):
        yield from range(self.nrows)


class TestExecuteTables(QueryBase):

    def setUp(self):
        super().setUp()
        self.ctx.tables['test'] = SimpleTable('test', 16)

    def execute(self, query):
        query = self.compile(query)
        return query()

    def test_null_table(self):
        self.assertEqual(
            self.execute("""SELECT 1 + 1 AS test FROM #"""),
            ((('test', int), ), [(2, )]))

    def test_simple_table(self):
        self.assertEqual(
            self.execute("""SELECT column FROM #test WHERE column < 2"""),
            ((('column', int), ), [(0, ), (1, )]))

    def test_simple_table_aggregation(self):
        self.assertEqual(
            self.execute("""SELECT sum(column) FROM #test GROUP BY column % 2"""),
            ((('sum(column)', int), ), [(56, ), (64, )]))


class TestInSubquery(QueryBase):
    data = """
      2000-01-01 open Income:Contributions:One
      2000-01-01 open Income:Contributions:Two
      2000-01-01 open Income:Contributions:Three
      2000-01-01 open Assets:Cash

      2024-10-25 * "One"
        Income:Contributions:One                            100.00 USD
        Assets:Cash

      2024-10-26 * "Two"
        Income:Contributions:Two                             10.00 USD
        Assets:Cash

      2024-10-27 * "Three"
        Income:Contributions:Three                           50.00 USD
        Assets:Cash

      2024-10-28 * "Three"
        Income:Contributions:Three                            1.00 USD
        Assets:Cash
    """

    def test_in_subquery(self):
        self.check_query(self.data, """
            SELECT
              account,
              date
            FROM #postings
            WHERE root(account, 2) = 'Income:Contributions'
            AND account IN (
              SELECT account
              FROM #postings
              GROUP BY account
              HAVING number(only('USD', sum(position))) > 50.0
            )
            ORDER BY date
            """,
            (('account', str), ('date', datetime.date)),
            [
            ('Income:Contributions:One',   datetime.date(2024, 10, 25)),
            ('Income:Contributions:Three', datetime.date(2024, 10, 27)),
            ('Income:Contributions:Three', datetime.date(2024, 10, 28)),
            ]
        )


class TestArrayOps(QueryBase):
    data = """
      1970-01-01 open Assets:Cash
      2000-01-01 open Expenses:One
      2000-01-02 open Expenses:Two
      2000-01-03 open Expenses:Three

      2025-01-01 * "One"
        Expenses:One                            100.00 USD
        Assets:Cash

      2025-01-02 * "Two"
        Expenses:Two                             10.00 USD
        Assets:Cash

      2025-01-03 * "Three"
        Expenses:Three                           50.00 USD
        Assets:Cash

      2025-01-03 * "Three"
        Expenses:Three                            1.00 USD
        Assets:Cash
    """

    def test_in_accounts_entries(self):
        self.check_query(self.data, """
            SELECT date
            FROM #entries
            WHERE 'Expenses:Two' IN accounts
            """,
            (('date', datetime.date),),
            [
            (datetime.date(2000, 1, 2),),
            (datetime.date(2025, 1, 2),),
            ]
        )

    def test_in_accounts_transactions(self):
        self.check_query(self.data, """
            SELECT date, narration
            FROM #transactions
            WHERE 'Expenses:Two' IN accounts
            """,
            (('date', datetime.date), ('narration', str)),
            [
            (datetime.date(2025, 1, 2), 'Two'),
            ]
        )

    def test_mathes_any_accounts_transactions(self):
        self.check_query(self.data, """
            SELECT date, narration
            FROM #transactions
            WHERE ':Two' ?~ ANY(accounts)
            """,
            (('date', datetime.date), ('narration', str)),
            [
            (datetime.date(2025, 1, 2), 'Two'),
            ]
        )

    def test_mathes_all_accounts_transactions(self):
        self.check_query(self.data, """
            SELECT date, narration
            FROM #transactions
            WHERE '(?i):two|:cash' ?~ ALL(accounts)
            """,
            (('date', datetime.date), ('narration', str)),
            [
            (datetime.date(2025, 1, 2), 'Two'),
            ]
        )


class TestArrayOpsSubquery(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = beanquery.connect('test:test?end=10')

    def check_query(self, query, expected_types, expected_rows):
        curs = self.conn.execute(query)
        self.assertEqual(tuple(expected_types), curs.description)
        result_rows = curs.fetchall()
        self.assertEqual(expected_rows, result_rows)

    def test_in_subquery(self):
        self.check_query(
            """
            SELECT 1 IN (SELECT x FROM #test) AS y FROM #
            """,
            (('y', bool),),
            [
            (True,),
            ]
        )
        self.check_query(
            """
            SELECT 1 IN (SELECT x FROM #test WHERE x % 2 = 0) AS y FROM #
            """,
            (('y', bool),),
            [
            (False,),
            ]
        )

    def test_any_subquery(self):
        self.check_query(
            """
            SELECT 1 = ANY (SELECT x FROM #test) AS y FROM #
            """,
            (('y', bool),),
            [
            (True,),
            ]
        )
        self.check_query(
            """
            SELECT 1 = ANY (SELECT x FROM #test WHERE x % 2 = 0) AS y FROM #
            """,
            (('y', bool),),
            [
            (False,),
            ]
        )

    def test_all_subquery(self):
        self.check_query(
            """
            SELECT 1 = ALL (SELECT x FROM #test) AS y FROM #
            """,
            (('y', bool),),
            [
            (False,),
            ]
        )
        self.check_query(
            """
            SELECT 1 = ALL (SELECT x FROM #test WHERE x = 1) AS y FROM #
            """,
            (('y', bool),),
            [
            (True,),
            ]
        )


class TestCreateTable(unittest.TestCase):

    def setUp(cls):
        cls.conn = beanquery.connect('')

    def test_create_table(self):
        curs = self.conn.execute('''CREATE TABLE abcd (a int, b bool, c str, d date)''')
        r = curs.fetchall()
        self.assertEqual(r, [])
        names = list(self.conn.tables['abcd'].columns.keys())
        self.assertEqual(names, ['a', 'b', 'c', 'd'])
        types = [column.dtype for column in self.conn.tables['abcd'].columns.values()]
        self.assertEqual(types, [int, bool, str, datetime.date])

    def test_create_table_as(self):
        self.conn.execute('''CREATE TABLE abc (a int, b bool, c str)''')
        self.conn.execute('''INSERT INTO abc (a, b, c) VALUES (1, true, '2025/01/01')''')
        curs = self.conn.execute('''CREATE TABLE test AS SELECT a, parse_date(c, '%Y/%m/%d') AS date FROM abc''')
        self.assertEqual(curs.fetchall(), [])
        names = list(self.conn.tables['test'].columns.keys())
        self.assertEqual(names, ['a', 'date'])
        types = [column.dtype for column in self.conn.tables['test'].columns.values()]
        self.assertEqual(types, [int, datetime.date])
        curs = self.conn.execute('''SELECT * FROM test''')
        self.assertEqual(curs.fetchall(), [(1, datetime.date(2025, 1, 1))])


class TestInsert(unittest.TestCase):

    def setUp(self):
        self.conn = beanquery.connect('')
        self.conn.execute('''CREATE TABLE abcd (a int, b bool, c str, d date)''')

    def test_insert_values(self):
        curs = self.conn.execute('''INSERT INTO abcd (a, b, c, d) VALUES (1, TRUE, 'one', 2025-01-01)''')
        self.assertEqual(self.conn.tables['abcd'].data[0], (1, True, 'one', datetime.date(2025, 1, 1)))
        self.assertEqual(curs.fetchall(), [])

    def test_insert_incomplete_row(self):
        curs = self.conn.execute('''INSERT INTO abcd (a, d) VALUES (1, 2025-01-01)''')
        self.assertEqual(self.conn.tables['abcd'].data[0], (1, None, None, datetime.date(2025, 1, 1)))
        self.assertEqual(curs.fetchall(), [])

    def test_insert_order(self):
        curs = self.conn.execute('''INSERT INTO abcd (a, d, c, b) VALUES (1, 2025-01-01, 'one', FALSE)''')
        self.assertEqual(self.conn.tables['abcd'].data[0], (1, False, 'one', datetime.date(2025, 1, 1)))
        self.assertEqual(curs.fetchall(), [])

    def test_insert_placeholders(self):
        values = (1, True, 'one', datetime.date(2025, 1, 1))
        curs = self.conn.execute('''INSERT INTO abcd (a, b, c, d) VALUES (%s, %s, %s, %s)''', values)
        self.assertEqual(self.conn.tables['abcd'].data[0], values)
        self.assertEqual(curs.fetchall(), [])


class TestCSVTable(unittest.TestCase):

    def setUp(self):
        self.conn = beanquery.connect('')

    @docfile
    def test_create_table(self, filename):
        '''\
        1234, 1, baz, 2025-01-01
        5678, 0, qux, 2025-01-02
        '''
        using = f'csv:{filename}'
        curs = self.conn.execute(f'''CREATE TABLE test (a int, b bool, c str, d date) USING {using!r}''')
        self.assertEqual(curs.fetchall(), [])
        curs = self.conn.execute('''SELECT * FROM test''')
        self.assertEqual(curs.fetchone(), (1234, True, 'baz', datetime.date(2025, 1, 1)))
        self.assertEqual(curs.fetchone(), (5678, False, 'qux', datetime.date(2025, 1, 2)))
        curs = self.conn.execute('''SELECT * FROM test''')
        self.assertEqual(len(curs.fetchall()), 2)

    @docfile
    def test_create_table_header(self, filename):
        '''\
        id
        1234
        5678
        '''
        using = f'csv:{filename}?header=1'
        curs = self.conn.execute(f'''CREATE TABLE test (a int) USING {using!r}''')
        self.assertEqual(curs.fetchall(), [])
        curs = self.conn.execute('''SELECT * FROM test''')
        self.assertEqual(len(curs.fetchall()), 2)

    @docfile
    def test_create_table_guess_types(self, filename):
        '''\
        id, name, check, date, value
        1234, one, true, 2025-01-01, 1.234
        5678, two, false, 2025-01-02, 5.678
        '''
        using = f'csv:{filename}'
        curs = self.conn.execute(f'''CREATE TABLE test USING {using!r}''')
        self.assertEqual(curs.fetchall(), [])
        names = list(self.conn.tables['test'].columns.keys())
        self.assertEqual(names, ['id', 'name', 'check', 'date', 'value'])
        types = [column.dtype for column in self.conn.tables['test'].columns.values()]
        self.assertEqual(types, [int, str, bool, datetime.date, Decimal])


class TestCSVSource(unittest.TestCase):

    @docfile
    def test_csv_source(self, filename):
        '''\
        id, name, check, date, value
        1234, one, true, 2025-01-01, 1.234
        '''
        conn = beanquery.connect(f'csv:{filename}?name=test')
        names = list(conn.tables['test'].columns.keys())
        self.assertEqual(names, ['id', 'name', 'check', 'date', 'value'])
        types = [column.dtype for column in conn.tables['test'].columns.values()]
        self.assertEqual(types, [int, str, bool, datetime.date, Decimal])



import unittest

from beanquery import types
from beanquery.sources import beancount


class TestName(unittest.TestCase):
    def test_transactions_names(self):
        table = beancount.TransactionsTable
        columns = {name: types.name(column.dtype) for name, column in table.columns.items()}
        self.assertEqual(columns, {
            'meta': 'metadata',
            'date': 'date',
            'flag': 'str',
            'payee': 'str',
            'narration': 'str',
            'tags': 'set',
            'links': 'set',
            'accounts': 'set[str]',
        })



import unittest
import sqlite3

import beanquery
from beanquery.sources import test


class APITests:
    def test_description(self):
        curs = self.conn.cursor()
        self.assertIsNone(curs.description)
        curs.execute(f'SELECT x FROM {self.table} WHERE x = 0')
        self.assertEqual([c[0] for c in curs.description], ['x'])
        column = curs.description[0]
        self.assertEqual(len(column), 7)

    def test_cursor_not_initialized(self):
        curs = self.conn.cursor()
        self.assertIsNone(curs.fetchone())
        self.assertEqual(curs.fetchmany(), [])
        self.assertEqual(curs.fetchall(), [])

    def test_cursor_fetchone(self):
        curs = self.conn.cursor()
        curs.execute(f'SELECT x FROM {self.table} WHERE x < 2')
        row = curs.fetchone()
        self.assertEqual(row, (0, ))
        row = curs.fetchone()
        self.assertEqual(row, (1, ))
        row = curs.fetchone()
        self.assertIsNone(row)

    def test_cursor_fetchall(self):
        curs = self.conn.cursor()
        curs.execute(f'SELECT x FROM {self.table} WHERE x < 2')
        rows = curs.fetchall()
        self.assertEqual(rows, [(0, ), (1, )])
        rows = curs.fetchall()
        self.assertEqual(rows, [])

    def test_cursor_fethmany(self):
        curs = self.conn.cursor()
        curs.execute(f'SELECT x FROM {self.table} WHERE x < 2')
        rows = curs.fetchmany()
        self.assertEqual(rows, [(0, )])
        rows = curs.fetchmany()
        self.assertEqual(rows, [(1, )])
        rows = curs.fetchmany()
        self.assertEqual(rows, [])

    def test_cursor_iterator(self):
        curs = self.conn.cursor()
        o = object()
        row = next(iter(curs), o)
        self.assertIs(row, o)
        curs = self.conn.cursor()
        curs.execute(f'SELECT x FROM {self.table} WHERE x < 2')
        iterator = iter(curs)
        row = next(iterator)
        self.assertEqual(row, (0, ))
        row = next(iterator)
        self.assertEqual(row, (1, ))
        row = next(iterator, o)
        self.assertIs(row, o)


class TestSQLite(APITests, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.table = 'test'
        cls.conn = sqlite3.connect(':memory:')
        curs = cls.conn.cursor()
        curs.execute('CREATE TABLE test (x int)')
        curs.executemany('INSERT INTO test VALUES (?)', [(i, ) for i in range(16)])

    @classmethod
    def tearDownClass(cls):
        cls.conn.close()


class TestBeanquery(APITests, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.table = '#test'
        cls.conn = beanquery.Connection()
        cls.conn.tables['test'] = test.Table(16)



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import atexit
import cmd
import importlib
import io
import itertools
import os
import pkgutil
import re
import shlex
import sys
import textwrap
import traceback
import warnings

from contextlib import nullcontext, suppress
from dataclasses import dataclass, asdict
from os import path

import click
import beancount

from beancount.parser import printer
from beancount.core import data
from beancount.utils import pager
from beancount.utils.misc_utils import get_screen_height

import beanquery

from beanquery import parser
from beanquery import query_compile
from beanquery import render
from beanquery import types
from beanquery.numberify import numberify_results

try:
    import readline
except ImportError:
    readline = None


HISTORY_FILENAME = '~/.config/beanquery/history'
INIT_FILENAME = '~/.config/beanquery/init'


class style:
    ERROR = '\033[31;1m'
    WARNING = '\033[31;1m'
    RESET = '\033[0m'

    ESCAPES = re.compile(r'\033\[[;?0-9]*[a-zA-Z]')

    @classmethod
    def strip(cls, x):
        return cls.ESCAPES.sub('', x)


def render_location(text, pos, endpos, lineno, indent, strip, out):
    length = endpos - pos
    lines = text.splitlines(True)
    for line in lines[:lineno]:
        pos -= len(line)
        if strip and not line.rstrip():
            continue
        strip = False
        out.append(indent + line.rstrip().expandtabs())
    out.append(indent + lines[lineno].rstrip().expandtabs())
    out.append(indent + ' ' * pos + '^' * length)


# FIXME: move the error formatting into the exception classes themselves
def render_exception(exc, indent='| ', strip=True):
    if isinstance(exc, (beanquery.CompilationError, beanquery.ParseError)) and exc.parseinfo:
        out = [str(exc)]
        pos = exc.parseinfo.pos
        endpos = exc.parseinfo.endpos
        lineno = exc.parseinfo.line
        render_location(exc.parseinfo.tokenizer.text, pos, endpos, lineno, indent, strip, out)
        return '\n'.join(out)
    return '\n' + traceback.format_exc()


FORMATS = {
    name: importlib.import_module('beanquery.render.' + name).render
    for finder, name, ispkg
    in pkgutil.iter_modules(render.__path__)
}


@dataclass
class Settings:
    boxed: bool = False
    expand: bool = False
    format: str = 'text'
    narrow: bool = True
    nullvalue: str = ''
    numberify: bool = False
    pager: bool = True
    spaced: bool = False
    unicode: bool = False

    def _parse_bool(self, value):
        if value in {True, False}:
            return value
        norm = value.strip().lower()
        if norm in {"1", "true", "t", "yes", "y", "on"}:
            return True
        if norm in {"0", "false", "f", "no", "n", "off"}:
            return False
        raise ValueError(f'"{value}" is not a valid boolean')

    def _parse_format(self, value):
        if value not in FORMATS:
            raise ValueError(f'"{value}" is not a valid format')
        return value

    def getstr(self, name):
        value = getattr(self, name)
        if isinstance(value, str):
            return repr(value)
        if isinstance(value, bool):
            return 'true' if value else 'false'
        return str(value)

    def setstr(self, name, value):
        vtype = type(getattr(self, name))
        parse = getattr(self, f'_parse_{name}', getattr(self, f'_parse_{vtype.__name__}', vtype))
        setattr(self, name, parse(value))

    def todict(self):
        return asdict(self)

    def __iter__(self):
        return iter(self.todict().keys())


class DispatchingShell(cmd.Cmd):
    """A usable convenient shell for interpreting commands, with history."""

    # Header for parsed commands.
    doc_header = "Shell utility commands (type help <topic>):"
    misc_header = "Beancount query commands:"

    def __init__(self, outfile, interactive, runinit, settings):
        """Create a shell with history.

        Args:
          outfile: An output file object to write communications to.
          interactive: A boolean, true if this serves an interactive tty.
          runinit: When true, execute the commands from the user init file.
          settings: The shell settings.
        """
        super().__init__()
        self.identchars += '.'
        self.outfile = outfile
        self.interactive = interactive
        self.settings = settings
        self.color = interactive and os.environ.get('TERM', 'dumb') != 'dumb'
        self.add_help()

        if interactive and readline is not None:

            # Setup completion on ``tab``. This needs to be done differently
            # depending on whether the ``readline`` backend is GNU readline or
            # libedit. This is handled automatically on Python 3.13 an
            # later. For Python 3.12 and earlier, the recommended way to check
            # for the string ``linedit`` in ``readline.__doc__``
            if 'libedit' in readline.__doc__:
                readline.parse_and_bind("bind ^I rl_complete")
            else:
                readline.parse_and_bind("tab: complete")

            # ``readline`` is used to complete command names, which are
            # strictly alphanumeric strings, and named query identifiers,
            # which may contain any ASCII character. To enable completion of
            # the latter, reduce the set of completion word delimiters to the
            # shell default. Notably remove ``-`` from the delimiters list
            # setup by Python.
            readline.set_completer_delims(" \t\n\"\\'`@$><=;|&{(")

            # Setup ``readline`` history handling.
            history_filepath = path.expanduser(HISTORY_FILENAME)
            os.makedirs(path.dirname(history_filepath), exist_ok=True)
            with suppress(FileNotFoundError):
                readline.read_history_file(history_filepath)
                readline.set_history_length(2048)
            atexit.register(readline.write_history_file, history_filepath)

        warnings.showwarning = self.warning

        if runinit:
            with suppress(FileNotFoundError):
                with open(path.expanduser(INIT_FILENAME)) as f:
                    for line in f:
                        self.onecmd(line)

    def echo(self, message, file=sys.stdout):
        if not self.color:
            message = style.strip(message)
        print(message, file=file)

    def error(self, message):
        self.echo(f'{style.ERROR}error:{style.RESET} {message}', file=sys.stderr)

    def warning(self, message, *args):
        self.echo(f'{style.WARNING}warning:{style.RESET} {message}', file=sys.stderr)

    def add_help(self):
        "Attach help functions for each of the parsed token handlers."
        for attrname, func in list(self.__class__.__dict__.items()):
            match = re.match('on_(.*)', attrname)
            if not match:
                continue
            command_name = match.group(1)
            setattr(self.__class__, f'help_{command_name.lower()}',
                    lambda _, fun=func: print(textwrap.dedent(fun.__doc__).strip(),
                                              file=self.outfile))

    @property
    def output(self):
        """Where to direct command output.

        When the output stream is connected to the standard output,
        and we are running interactively, use an indirection that can
        send the output to a pager when the number of lines emitted is
        greater than a threshold.

        Returns:
          A context manager that returns a file descriptor on enter.

        """
        if self.outfile is sys.stdout:
            if self.interactive and self.settings.pager:
                return pager.ConditionalPager(None, minlines=get_screen_height())
            return pager.flush_only(sys.stdout)
        return nullcontext(self.outfile)

    def cmdloop(self, intro=None):
        """Override cmdloop to handle keyboard interrupts."""
        while True:
            try:
                super().cmdloop(intro)
                break
            except KeyboardInterrupt:
                print('\n(interrupted)', file=sys.stderr)
            except Exception as exc:
                self.error(render_exception(exc))

    def parseline(self, line):
        cmd, arg, line = super().parseline(line)
        if not cmd:
            return cmd, arg, line
        if cmd.startswith('.'):
            cmd = cmd[1:]
        if cmd == 'EOF':
            line = '.EOF'
        return cmd, arg, line

    def onecmd(self, line):
        cmd, arg, line = self.parseline(line)
        if not cmd:
            return
        if not line.startswith('.'):
            cmd = cmd.lower()
            if cmd not in {'clear', 'errors', 'exit', 'help', 'history', 'parse', 'quit', 'run', 'set'}:
                return self.execute(line)
            warnings.warn(f'commands without "." prefix are deprecated. use ".{cmd}" instead', stacklevel=0)
        func = getattr(self, 'do_' + cmd, None)
        if func is not None:
            return func(arg)
        self.error(f'unknown command "{cmd}"')

    def completenames(self, text, *ignored):
        if text.startswith('.'):
            dotext = 'do_' + text[1:]
            return ['.' + a[3:] for a in self.get_names() if a.startswith(dotext)]

    def do_help(self, arg):
        """List available commands with "help" or detailed help with "help cmd"."""
        super().do_help(arg.lower())

    def do_history(self, arg):
        """Print the command-line history."""
        if readline is not None:
            num_entries = readline.get_current_history_length()
            try:
                max_entries = int(arg)
                start = max(0, num_entries - max_entries)
            except ValueError:
                start = 0
            for index in range(start, num_entries):
                line = readline.get_history_item(index + 1)
                print(line, file=self.outfile)

    def do_clear(self, arg):
        """Clear the command-line history."""
        if readline is not None:
            readline.clear_history()

    def do_set(self, arg):
        """Set shell settings variables."""
        if not arg:
            for name in self.settings:
                value = self.settings.getstr(name)
                print(f'{name}: {value}', file=self.outfile)
        else:
            components = shlex.split(arg)
            name = components[0]
            if len(components) == 1:
                try:
                    value = self.settings.getstr(name)
                    print(f'{name}: {value}', file=self.outfile)
                except AttributeError:
                    self.error(f'variable "{name}" does not exist')
            elif len(components) == 2:
                value = components[1]
                try:
                    self.settings.setstr(name, value)
                except ValueError as ex:
                    self.error(str(ex))
                except AttributeError:
                    self.error(f'variable "{name}" does not exist')
            else:
                self.error('invalid number of arguments')

    def complete_set(self, text, _line, _begidx, _endidx):
        return [name for name in self.settings if name.startswith(text)]

    def do_format(self, arg):
        """Set output format to FRMT."""
        if not arg:
            frmt = self.settings.getstr('format')
            print(f'format: {frmt}', file=self.outfile)
        else:
            value, *others = shlex.split(arg)
            if others:
                self.error('invalid number of arguments')
            try:
                self.settings.setstr('format', value)
            except ValueError as ex:
                self.error(str(ex))

    def complete_format(self, text, _line, _begidx, _endidx):
        return [frmt for frmt in FORMATS if frmt.startswith(text)]

    def do_output(self, arg):
        """Send output to FILE or stdout if FILE is omitted."""
        if self.outfile is not sys.stdout:
            self.outfile.close()
        self.outfile = open(arg, "w") if arg else open(sys.stdout)

    def do_parse(self, arg):
        """Run the parser on the following command and print the output."""
        print(self.parse(arg).tosexp())

    def execute(self, query, **kwargs):
        raise NotImplementedError

    def do_exit(self, arg):
        """Exit the command interpreter."""
        return True

    do_quit = do_exit

    def do_EOF(self, arg):
        """Exit the command interpreter."""
        print('exit', file=self.outfile)
        return self.do_exit(arg)


class BQLShell(DispatchingShell):
    """An interactive shell interpreter for the Beancount query language."""
    prompt = 'beanquery> '

    def __init__(self, source, outfile, interactive=False, runinit=False, format='text', numberify=False, errors=True):
        settings = Settings(format=format, numberify=numberify)
        super().__init__(outfile, interactive, runinit, settings)
        self.context = beanquery.connect(None)
        self.source = source
        self.show_load_errors = errors
        self.queries = {}
        self.do_reload()

    def parse(self, line, default_close_date=None, **kwargs):
        statement = self.context.parse(line)
        if (isinstance(statement, parser.ast.Select) and
            isinstance(statement.from_clause, parser.ast.From) and
            not statement.from_clause.close):
            statement.from_clause.close = default_close_date
        return statement

    def execute(self, query, **kwargs):
        statement = self.parse(query, **kwargs)
        name = type(statement).__name__
        method = getattr(self, f'on_{name}', None)
        if method is not None:
            return method(statement)
        return self.on_Select(statement)

    def do_reload(self, arg=None):
        "Reload the Beancount input file."
        self.context.errors.clear()
        self.context.options.clear()
        if self.source:
            self.context.attach(self.source)
        if self.source.startswith('beancount:'):
            table = self.context.tables['entries']
            self._extract_queries(table.entries)
            if self.context.errors and self.show_load_errors:
                printer.print_errors(self.context.errors, file=sys.stderr)
            if self.interactive:
                print_statistics(table.entries, table.options, self.context.errors, self.outfile)

    def _extract_queries(self, entries):
        self.queries = {}
        for entry in entries:
            if isinstance(entry, data.Query):
                x = self.queries.setdefault(entry.name, entry)
                if x is not entry:
                    warnings.warn(f'duplicate query name "{entry.name}"', stacklevel=0)

    def do_errors(self, arg):
        "Print the errors that occurred during Beancount input file parsing."
        if self.context.errors:
            printer.print_errors(self.context.errors)
        else:
            print('(no errors)', file=self.outfile)

    def do_run(self, arg):
        "Run a named query defined in the Beancount input file."

        arg = arg.rstrip('; \t')
        if not arg:
            # List the available queries.
            if self.queries:
                print('\n'.join(name for name in sorted(self.queries)))
            return

        if arg == "*":
            # Execute all.
            for name, query in sorted(self.queries.items()):
                print(f'{name}:')
                self.execute(query.query_string, default_close_date=query.date)
                print()
                print()
            return

        name, *args = shlex.split(arg)
        if args:
            self.error('too many arguments for "run" command')
            return

        query = self.queries.get(name)
        if not query:
            self.error(f'query "{name}" not found')
            return
        self.execute(query.query_string, default_close_date=query.date)

    def complete_run(self, text, line, begidx, endidx):
        return [name for name in self.queries if name.startswith(text)]

    def do_tables(self, arg):
        """List tables."""
        print('\n'.join(sorted(filter(None, self.context.tables.keys()))), file=self.outfile)

    def do_describe(self, arg):
        """Describe table or structured type."""
        def describe(obj):
            return '\n'.join(f'  {name} ({types.name(column.dtype)})' for name, column in obj.columns.items())
        names = shlex.split(arg)
        for name in names:
            table = self.context.tables.get(name)
            if table:
                print(f'table {name}:', file=self.outfile)
                print(describe(table), file=self.outfile)
            datatype = types.TYPES.get(name)
            if datatype:
                print(f'structured type {name}:', file=self.outfile)
                print(describe(datatype), file=self.outfile)

    def complete_describe(self, text, line, begidx, endidx):
        names = itertools.chain(self.context.tables.keys(), types.TYPES.keys())
        return [name for name in names if name and name.startswith(text)]

    def do_explain(self, arg):
        """Compile and print a compiled statement for debugging."""

        p = lambda x: print(x, file=self.outfile)

        statement = self.context.parse(arg)
        p('parsed statement')
        p('----------------')
        p(textwrap.indent(statement.tosexp(), '  '))
        p('')

        query = self.context.compile(statement)
        p('compiled query')
        p('--------------')
        p(f'  {query}')
        p('')

        if not isinstance(query, query_compile.EvalNode):
            return

        p('targets')
        p('-------')
        for target in query.c_targets:
            name = target.name or ''
            datatype = types.name(target.c_expr.dtype)
            if target.is_aggregate:
                datatype += ', aggregate'
            p(f'  {name}: {datatype}')

    def on_Print(self, statement):
        """
        Print entries in Beancount format.

        The general form of a PRINT statement includes an SQL-like FROM
        selector:

           PRINT [FROM <from_expr> ...]

        Where:

          from_expr: A logical expression that matches on the attributes of
            the directives. See SELECT command for details (this FROM expression
            supports all the same expressions including its OPEN, CLOSE and
            CLEAR operations).

        """
        frmt = self.settings.format
        try:
            self.settings.format = 'beancount'
            self.on_Select(statement)
        finally:
            self.settings.format = frmt

    def on_Select(self, statement):
        """
        Extract data from a query on the postings.

        The general form of a SELECT statement loosely follows SQL syntax, with
        some mild and idiomatic extensions:

           SELECT [DISTINCT] [<targets>|*]
           [FROM <from_expr> [OPEN ON <date>] [CLOSE [ON <date>]] [CLEAR]]
           [WHERE <where_expr>]
           [GROUP BY <groups>]
           [ORDER BY <groups> [ASC|DESC]]
           [LIMIT num]

        Where:

          targets: A list of desired output attributes from the postings, and
            expressions on them. Some of the attributes of the parent transaction
            directive are made available in this context as well. Simple functions
            (that return a single value per row) and aggregation functions (that
            return a single value per group) are available. For the complete
            list of supported columns and functions, see help on "targets".
            You can also provide a wildcard here, which will select a reasonable
            default set of columns for rendering a journal.

          from_expr: A logical expression that matches on the attributes of
            the directives (not postings). This allows you to select a subset of
            transactions, so the accounting equation is respected for balance
            reports. For the complete list of supported columns and functions,
            see help on "from".

          where_expr: A logical expression that matches on the attributes of
            postings. The available columns are similar to those in the targets
            clause, without the aggregation functions.

          OPEN clause: replace all the transactions before the given date by
            summarizing entries and transfer Income and Expenses balances to
            Equity.

          CLOSE clause: Remove all the transactions after the given date and

          CLEAR: Transfer final Income and Expenses balances to Equity.

        """
        cursor = self.context.execute(statement)
        desc = cursor.description
        rows = cursor.fetchall()
        dcontext = self.context.options.get('dcontext', None)

        if self.settings.numberify:
            desc, rows = numberify_results(desc, rows, dcontext.build())

        with self.output as out:
            render = FORMATS.get(self.settings.format)
            if render is not None:
                return render(desc, rows, out, dcontext=dcontext, **self.settings.todict())
            raise NotImplementedError

    def on_Journal(self, journal):
        """
        Select a journal of some subset of postings. This command is a
        convenience and converts into an equivalent Select statement, designed
        to extract the most sensible list of columns for the register of a list
        of entries as a table.

        The general form of a JOURNAL statement loosely follows SQL syntax:

           JOURNAL <account-regexp> [FROM_CLAUSE]

        See the SELECT query help for more details on the FROM clause.
        """
        return self.on_Select(journal)

    def on_Balances(self, balance):
        """
        Select balances of some subset of postings. This command is a
        convenience and converts into an equivalent Select statement, designed
        to extract the most sensible list of columns for the register of a list
        of entries as a table.

        The general form of a JOURNAL statement loosely follows SQL syntax:

           BALANCE [FROM_CLAUSE]

        See the SELECT query help for more details on the FROM clause.
        """
        return self.on_Select(balance)

    def help_targets(self):
        template = textwrap.dedent("""

          The list of comma-separated target expressions may consist of columns,
          simple functions and aggregate functions. If you use any aggregate
          function, you must also provide a GROUP-BY clause.

          Columns
          -------

          {columns}

          Functions
          ---------

          {functions}

          Aggregate functions
          -------------------

          {aggregates}

        """)
        print(template.format(**_describe(self.context.tables['postings'],
                                          query_compile.FUNCTIONS)), file=self.outfile)

    def help_from(self):
        template = textwrap.dedent("""

          A logical expression that consist of columns on directives (mostly
          transactions) and simple functions.

          Columns
          -------

          {columns}

          Functions
          ---------

          {functions}

        """)
        print(template.format(**_describe(self.context.tables['entries'],
                                          query_compile.FUNCTIONS)),
              file=self.outfile)

    def help_where(self):
        template = textwrap.dedent("""

          A logical expression that consist of columns on postings and simple
          functions.

          Columns
          -------

          {columns}

          Functions
          ---------

          {functions}

        """)
        print(template.format(**_describe(self.context.tables['postings'],
                                          query_compile.FUNCTIONS)), file=self.outfile)


def _describe_columns(columns):
    out = io.StringIO()
    wrapper = textwrap.TextWrapper(initial_indent='  ', subsequent_indent='  ', width=80)
    for name, column in columns.items():
        print(f'{name}: {types.name(column.dtype)}', file=out)
        print(wrapper.fill(re.sub(r'[ \n\t]+', ' ', column.__doc__ or '')), file=out)
        print(file=out)
    return out.getvalue().rstrip()


def _describe_functions(functions, aggregates=False):
    entries = []
    for name, funcs in functions.items():
        if aggregates != issubclass(funcs[0], query_compile.EvalAggregator):
            continue
        name = name.lower()
        for func in funcs:
            args = ', '.join(types.name(d) for d in func.__intypes__)
            doc = re.sub(r'[ \n\t]+', ' ', func.__doc__ or '')
            entries.append((name, doc, args))
    entries.sort()
    out = io.StringIO()
    wrapper = textwrap.TextWrapper(initial_indent='  ', subsequent_indent='  ', width=80)
    for key, entries in itertools.groupby(entries, key=lambda x: x[:2]):  # noqa: B020
        for name, doc, args in entries:
            print(f'{name}({args})', file=out)
        print(wrapper.fill(doc), file=out)
        print(file=out)
    return out.getvalue().rstrip()


def _describe(table, functions):
    return dict(
        columns=_describe_columns(table.columns),
        functions=_describe_functions(functions, aggregates=False),
        aggregates=_describe_functions(functions, aggregates=True))


def summary_statistics(entries):
    """Calculate basic summary statistics to output a brief welcome message.

    Args:
      entries: A list of directives.
    Returns:
      A tuple of three integers, the total number of directives parsed, the total number
      of transactions and the total number of postings there in.
    """
    num_directives = len(entries)
    num_transactions = 0
    num_postings = 0
    for entry in entries:
        if isinstance(entry, data.Transaction):
            num_transactions += 1
            num_postings += len(entry.postings)
    return num_directives, num_transactions, num_postings


def print_statistics(entries, options, errors, outfile):
    """Print summary statistics to stdout.

    Args:
      entries: A list of directives.
      options: An options map. as produced by the parser.
      outfile: A file object to write to.
    """
    num_directives, num_transactions, num_postings = summary_statistics(entries)
    if 'title' in options:
        print(f'''Input file: "{options['title']}"''', file=outfile)
    print(f"Ready with {num_directives} directives "
          f"({num_postings} postings in {num_transactions} transactions, "
          f"{len(errors)} validation errors)",
          file=outfile)


@click.command()
@click.argument('filename')
@click.argument('query', nargs=-1)
@click.option('--output', '-o', type=click.File('w'), default='-',
              help="Output filename.")
@click.option('--format', '-f', type=click.Choice(FORMATS.keys()), default='text',
              help="Output format.")
@click.option('--numberify', '-m', is_flag=True,
              help="Numberify the output, removing the currencies.")
@click.option('--no-errors', '-q', is_flag=True,
              help="Do not report ledger validation errors on load.")
@click.version_option('', message=f'beanquery {beanquery.__version__}, beancount {beancount.__version__}')
def main(filename, query, numberify, format, output, no_errors):
    """An interactive interpreter for the Beancount Query Language.

    Load Beancount ledger FILENAME and run Beancount Query Language
    QUERY on it, if specified, or drop into the interactive shell. If
    not explicitly set with the dedicated option, the output format is
    inferred from the output file name, if specified.

    """
    # Prepend ``beancount:`` to DSNs not having a ``scheme:`` prefix.
    # Take into account file paths starting with a single letter
    # Windows-style drive designations, which can be spelled with
    # forward or backward slashes.
    source = filename if re.match('[a-z]{2,}:',filename) else 'beancount:' + filename

    # Create the shell.
    interactive = sys.stdin.isatty() and not query
    shell = BQLShell(source, output, interactive, True, format, numberify, not no_errors)

    # Run interactively if we're a TTY and no query is supplied.
    if interactive:
        warnings.filterwarnings('always')
        shell.cmdloop()
    else:
        # Run in batch mode (Non-interactive).
        if query:
            # We have a query to run.
            query = ' '.join(query)
        else:
            # If we have no query and we're not a TTY, read the BQL command from
            # standard input.
            query = sys.stdin.read()

        shell.onecmd(query)



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import unittest
from decimal import Decimal

from beancount.core.number import D
from beancount.parser import parser
from beancount import loader
from beanquery import query_compile as qc
from beanquery import query_env as qe
from beanquery import query


class TestCompileDataTypes(unittest.TestCase):

    def test_compile_Length(self):
        c_length = qe.Function('length', [qc.EvalConstant('testing')])
        self.assertEqual(int, c_length.dtype)

    def test_compile_Sum(self):
        c_sum = qe.SumInt(None, [qc.EvalConstant(17)])
        self.assertEqual(int, c_sum.dtype)
        c_sum = qe.SumDecimal(None, [qc.EvalConstant(D('17.'))])
        self.assertEqual(Decimal, c_sum.dtype)

    def test_compile_Count(self):
        c_count = qe.Count(None, [qc.EvalConstant(17)])
        self.assertEqual(int, c_count.dtype)

    def test_compile_First(self):
        c_first = qe.First(None, [qc.EvalConstant(17.)])
        self.assertEqual(float, c_first.dtype)

    def test_compile_Last(self):
        c_last = qe.Last(None, [qc.EvalConstant(17.)])
        self.assertEqual(float, c_last.dtype)


class TestEnv(unittest.TestCase):

    @parser.parse_doc()
    def test_GrepN(self, entries, _, options_map):
        """
        2016-11-20 * "prev match in context next"
          Assets:Banking          1 USD
        """
        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT GREPN("in", narration, 0) as m
        ''')
        self.assertEqual([('in',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT GREPN("match (.*) context", narration, 1) as m
        ''')
        self.assertEqual([('in',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT GREPN("(.*) in (.*)", narration, 2) as m
        ''')
        self.assertEqual([('context next',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT GREPN("ab(at)hing", "abathing", 1) as m
        ''')
        self.assertEqual([('at',)], rrows)

    @parser.parse_doc()
    def test_Subst(self, entries, _, options_map):
        """
        2016-11-20 * "I love candy"
          Assets:Banking       -1 USD

        2016-11-21 * "Buy thing thing"
          Assets:Cash          -1 USD
        """
        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT SUBST("[Cc]andy", "carrots", narration) as m where date = 2016-11-20
        ''')
        self.assertEqual([('I love carrots',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT SUBST("thing", "t", narration) as m where date = 2016-11-21
        ''')
        self.assertEqual([('Buy t t',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT SUBST("random", "t", narration) as m where date = 2016-11-21
        ''')
        self.assertEqual([('Buy thing thing',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT SUBST("(love)", "\\1 \\1", narration) as m where date = 2016-11-20
        ''')
        self.assertEqual([('I love love candy',)], rrows)

        rtypes, rrows = query.run_query(entries, options_map, '''
          SELECT SUBST("Assets:.*", "Savings", account) as a, str(sum(position)) as p
        ''')
        self.assertEqual([('Savings', '(-2 USD)')], rrows)

    @parser.parse_doc()
    def test_Date(self, entries, _, options_map):
        """
        2016-11-20 * "ok"
          Assets:Banking          1 USD
        """
        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date(2020, 1, 2) as m')
        self.assertEqual([(datetime.date(2020, 1, 2),)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date(year, month, 1) as m')
        self.assertEqual([(datetime.date(2016, 11, 1),)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date(2020, 2, 32) as m')
        self.assertEqual([(None,)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date("2020-01-02") as m')
        self.assertEqual([(datetime.date(2020, 1, 2),)], rrows)


    @parser.parse_doc()
    def test_DateDiffAdjust(self, entries, _, options_map):
        """
        2016-11-20 * "ok"
          Assets:Banking          -1 STOCK { 5 USD, 2016-10-30 }
        """
        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date_diff(date, cost_date) as m')
        self.assertEqual([(21,)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date_diff(cost_date, date) as m')
        self.assertEqual([(-21,)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date_add(date, 1) as m')
        self.assertEqual([(datetime.date(2016, 11, 21),)], rrows)

        rtypes, rrows = query.run_query(entries, options_map,
                                        'SELECT date_add(date, -1) as m')
        self.assertEqual([(datetime.date(2016, 11, 19),)], rrows)

    def test_func_meta(self):
        # use the loader to have the pad transaction inserted
        entries, _, options = loader.load_string('''
          2019-01-01 open Assets:Main USD
          2019-01-01 open Assets:Other USD
          2019-01-14 * "Test"
            entry: 1
            both: 3
            Assets:Main  100.00 USD
              post: 2
              both: 3
            Assets:Other
              post: 4
          2019-01-15 pad Assets:Main Assets:Other
          2019-01-16 balance Assets:Main 1000.00 USD
        ''', dedent=True)
        rtypes, rrows = query.run_query(entries, options, '''
          SELECT
            entry_meta('post'),
            meta('post'),
            any_meta('post'),
            entry_meta('entry'),
            meta('entry'),
            any_meta('entry'),
            any_meta('both')
        ''')
        self.assertEqual([
            (None, D(2), D(2), D(1), None, D(1), D(3)),
            (None, D(4), D(4), D(1), None, D(1), D(3)),
            # postings from pad directive
            (None, None, None, None, None, None, None),
            (None, None, None, None, None, None, None),
        ], rrows)


if __name__ == '__main__':
    unittest.main()



"""
Exceptions hierarchy defined by the DB-API:

    Exception
      Warning
      Error
        InterfaceError
        DatabaseError
          DataError
          OperationalError
          IntegrityError
          InternalError
          ProgrammingError
          NotSupportedError
"""


class Warning(Exception):
    """Exception raised for important warnings."""

    __module__ = 'beanquery'


class Error(Exception):
    """Base exception for all errors."""

    __module__ = 'beanquery'


class InterfaceError(Error):
    """An error related to the database interface rather than the database itself."""

    __module__ = 'beanquery'


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""

    __module__ = 'beanquery'


class DataError(DatabaseError):
    """An error caused by problems with the processed data."""

    __module__ = 'beanquery'


class OperationalError(DatabaseError):
    """An error related to the database's operation."""

    __module__ = 'beanquery'


class IntegrityError(DatabaseError):
    """An error caused when the relational integrity of the database is affected."""

    __module__ = 'beanquery'


class InternalError(DatabaseError):
    """An error generated when the database encounters an internal error."""

    __module__ = 'beanquery'


class ProgrammingError(DatabaseError):
    """Exception raised for programming errors."""

    __module__ = 'beanquery'


class NotSupportedError(DatabaseError):
    """A method or database API was used which is not supported by the database."""

    __module__ = 'beanquery'



from operator import attrgetter
from typing import Sequence

from . import types
from . import parser
from . import compiler


class Column(Sequence):
    __module__ = 'beanquery'

    def __init__(self, name, datatype):
        self._name = name
        self._type = datatype

    _vars = tuple(attrgetter(name) for name in 'name type_code display_size internal_size precision scale null_ok'.split())

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return tuple(self) == tuple(other)
        if isinstance(other, tuple):
            # Used in tests.
            return (self._name, self._type) == other
        return NotImplemented

    def __repr__(self):
        return f'{self.__module__}.{self.__class__.__name__}({self._name!r}, {types.name(self._type)})'

    def __len__(self):
        return 7

    def __getitem__(self, key):
        if isinstance(key, slice):
            return tuple(getter(self) for getter in self._vars(key))
        return self._vars[key](self)

    @property
    def name(self):
        return self._name

    @property
    def datatype(self):
        # Extension to the DB-API.
        return self._type

    @property
    def type_code(self):
        # The DB-API specification is vague on this point, but other
        # database connection libraries expose this as an int. It does
        # not make much sense to keep a mapping between int type code
        # and actual types, thus just return the hash of the type
        # object.
        return hash(self._type)

    @property
    def display_size(self):
        return None

    @property
    def internal_size(self):
        return None

    @property
    def precision(self):
        return None

    @property
    def scale(self):
        return None

    @property
    def null_ok(self):
        return None


class Cursor:
    def __init__(self, connection):
        self._context = connection
        self._description = None
        self._rows = None
        self._pos = 0
        self.arraysize = 1

    @property
    def connection(self):
        return self._context

    def execute(self, query, params=None):
        if not isinstance(query, parser.ast.Node):
            query = parser.parse(query)
        query = compiler.compile(self._context, query, params)
        description, rows = query()
        self._description = description
        self._rows = rows
        self._pos = 0
        return self

    def executemany(self, query, params=None):
        query = parser.parse(query)
        for p in params:
            self.execute(query, p)

    @property
    def description(self):
        return self._description

    @property
    def rowcount(self):
        return len(self._rows) if self._rows is not None else -1

    @property
    def rownumber(self):
        return self._pos

    def fetchone(self):
        # This implementation pops items from the front of the results
        # rows list and is thus not efficient, especially for large
        # results sets.
        if self._rows is None or not len(self._rows):
            return None
        self._pos += 1
        return self._rows.pop(0)

    def fetchmany(self, size=None):
        if self._rows is None:
            return []
        n = size if size is not None else self.arraysize
        rows = self._rows[:n]
        self._rows = self._rows[n:]
        self._pos += len(rows)
        return rows

    def fetchall(self):
        if self._rows is None:
            return []
        rows = self._rows
        self._rows = []
        self._pos += len(rows)
        return rows

    def close(self):
        # Required by the DB-API.
        pass

    def setinputsizes(self, sizes):
        # Required by the DB-API.
        pass

    def setoutputsize(self, size, column=None):
        # Required by the DB-API.
        pass

    def __iter__(self):
        return iter(self._rows if self._rows is not None else [])



from __future__ import annotations

import dataclasses
import datetime
import enum
import sys
import textwrap
import typing


if typing.TYPE_CHECKING:
    from typing import Any, Optional, Union


def _indent(text):
    return textwrap.indent(text, '  ')


def _fields(node):
    for field in dataclasses.fields(node):
        if field.repr:
            yield field.name, getattr(node, field.name)


def tosexp(node):
    if isinstance(node, Node):
        return f'({node.__class__.__name__.lower()}\n' + _indent(
            '\n'.join(f'{name.replace("_", "-")}: {tosexp(value)}'
                      for name, value in _fields(node) if value is not None) + ')')
    if isinstance(node, list):
        return '(\n' + _indent('\n'.join(tosexp(i) for i in node)) + ')'
    if isinstance(node, enum.Enum):
        return node.name.lower()
    return repr(node)


def walk(node):
    if isinstance(node, Node):
        for name, child in _fields(node):
            yield from walk(child)
        yield node
    if isinstance(node, list):
        for child in node:
            yield from walk(child)


class Node:
    """Base class for BQL AST nodes."""
    __slots__ = ()
    parseinfo = None

    @property
    def text(self):
        if not self.parseinfo:
            return None
        text = self.parseinfo.tokenizer.text
        return text[self.parseinfo.pos:self.parseinfo.endpos]

    def tosexp(self):
        return tosexp(self)

    def walk(self):
        return walk(self)


def node(name, fields):
    """Manufacture an AST node class."""

    return dataclasses.make_dataclass(
        name,
        [*fields.split(), ('parseinfo', None, dataclasses.field(default=None, compare=False, repr=False))],
        bases=(Node,),
        **({'slots': True} if sys.version_info[:2] >= (3, 10) else {}))


# A 'select' query action.
#
# Attributes:
#   targets: Either a single 'Asterisk' instance of a list of 'Target'
#     instances.
#   from_clause: An instance of 'From', or None if absent.
#   where_clause: A root expression node, or None if absent.
#   group_by: An instance of 'GroupBy', or None if absent.
#   order_by: An instance of 'OrderBy', or None if absent.
#   pivot_by: An instance of 'PivotBy', or None if absent.
#   limit: An integer, or None is absent.
#   distinct: A boolean value (True), or None if absent.
Select = node('Select', 'targets from_clause where_clause group_by order_by pivot_by limit distinct')

# A select query that produces final balances for accounts.
# This is equivalent to
#
#   SELECT account, sum(position)
#   FROM ...
#   WHERE ...
#   GROUP BY account
#
# Attributes:
#   summary_func: A method on an inventory to call on the position column.
#     May be to extract units, value at cost, etc.
#   from_clause: An instance of 'From', or None if absent.
Balances = node('Balances', 'summary_func from_clause where_clause')

# A select query that produces a journal of postings.
# This is equivalent to
#
#   SELECT date, flag, payee, narration, ...  FROM <from_clause>
#   WHERE account = <account>
#
# Attributes:
#   account: A string, the name of the account to restrict to.
#   summary_func: A method on an inventory to call on the position column.
#     May be to extract units, value at cost, etc.
#   from_clause: An instance of 'From', or None if absent.
Journal = node('Journal', 'account summary_func from_clause')

# A query that will simply print the selected entries in Beancount format.
#
# Attributes:
#   from_clause: An instance of 'From', or None if absent.
Print = node('Print', 'from_clause')

# A parsed SELECT column or target.
#
# Attributes:
#   expression: A tree of expression nodes from the parser.
#   name: A string, the given name of the target (given by "AS <name>").
Target = node('Target', 'expression name')

# A placeholder in SELECT * or COUNT(*) constructs.
Asterisk = node('Asterisk', '')

# A FROM clause.
#
# Attributes:
#   expression: A tree of expression nodes from the parser.
#   close: A CLOSE clause, either None if absent, a boolean if the clause
#     was present by no date was provided, or a datetime.date instance if
#     a date was provided.
@dataclasses.dataclass(**({'slots': True} if sys.version_info[:2] >= (3, 10) else {}))
class From(Node):
    expression: Optional[Node] = None
    open: Optional[datetime.date] = None
    close: Optional[Union[datetime.date, bool]] = None
    clear: Optional[bool] = None
    parseinfo: Any = dataclasses.field(default=None, compare=False, repr=False)

# A GROUP BY clause.
#
# Attributes:
#   columns: A list of group-by expressions, simple Column() or otherwise.
#   having: An expression tree for the optional HAVING clause, or None.
GroupBy = node('GroupBy', 'columns having')

# An ORDER BY clause.
#
# Attributes:
#   column: order-by expression, simple Column() or otherwise.
#   ordering: The sort order as an Ordering enum value.
OrderBy = node('OrderBy', 'column ordering')

class Ordering(enum.IntEnum):
    # The enum values are chosen in this way to be able to use them
    # directly as the reverse parameter to the list sort() method.
    ASC = 0
    DESC = 1

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"

# An PIVOT BY clause.
#
# Attributes:
#   columns: A list of group-by expressions, simple Column() or otherwise.
PivotBy = node('PivotBy', 'columns')

# A reference to a table.
#
# Attributes:
#   name: The table name.
Table = node('Table', 'name')

# A reference to a column.
#
# Attributes:
#   name: A string, the name of the column to access.
Column = node('Column', 'name')

# A function call.
#
# Attributes:
#   fname: A string, the name of the function.
#   operands: A list of other expressions, the arguments of the function to
#     evaluate. This is possibly an empty list.
Function = node('Function', 'fname operands')

Attribute = node('Attribute', 'operand name')

Subscript = node('Subscript', 'operand key')

# A constant node.
#
# Attributes:
#   value: The constant value this represents.
Constant = node('Constant', 'value')

# A query parameter placeholder.
#
# Attributes:
#   name: The placeholder name
Placeholder = node('Placeholder', 'name')

# Base class for unary operators.
#
# Attributes:
#   operand: An expression, the operand of the operator.
UnaryOp = node('UnaryOp', 'operand')

# Base class for binary operators.
#
# Attributes:
#   left: An expression, the left operand.
#   right: An expression, the right operand.
BinaryOp = node('BinaryOp', 'left right')

# Base class for boolean operators.
BoolOp = node('BoolOp', 'args')

# Between
Between = node('Between', 'operand lower upper')

# Negation operator.
class Not(UnaryOp):
    __slots__ = ()

class IsNull(UnaryOp):
    __slots__ = ()

class IsNotNull(UnaryOp):
    __slots__ = ()


# Boolean operators.

class And(BoolOp):
    __slots__ = ()

class Or(BoolOp):
    __slots__ = ()


# Equality and inequality comparison operators.

class Equal(BinaryOp):
    __slots__ = ()

class NotEqual(BinaryOp):
    __slots__ = ()

class Greater(BinaryOp):
    __slots__ = ()

class GreaterEq(BinaryOp):
    __slots__ = ()

class Less(BinaryOp):
    __slots__ = ()

class LessEq(BinaryOp):
    __slots__ = ()


# Regular expression match operator.

class Match(BinaryOp):
    __slots__ = ()


class NotMatch(BinaryOp):
    __slots__ = ()


class Matches(BinaryOp):
    __slots__ = ()


# Membership operators.

class In(BinaryOp):
    __slots__ = ()

class NotIn(BinaryOp):
    __slots__ = ()


# Arithmetic operators.

class Neg(UnaryOp):
    __slots__ = ()

class Mul(BinaryOp):
    __slots__ = ()

class Div(BinaryOp):
    __slots__ = ()

class Mod(BinaryOp):
    __slots__ = ()

class Add(BinaryOp):
    __slots__ = ()

class Sub(BinaryOp):
    __slots__ = ()


Any = node('Any', 'left op right')
All = node('All', 'left op right')


CreateTable = node('CreateTable', 'name columns using query')

Insert = node('Insert', 'table columns values')



__copyright__ = "Copyright (C) 2015-2017  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beancount import loader
from beanquery import query


class TestSimple(unittest.TestCase):

    @loader.load_doc()
    def test_run_query(self, entries, _, options):
        """
        2022-01-01 open Assets:Checking         USD
        2022-01-01 open Income:ACME             USD
        2022-01-01 open Expenses:Taxes:Federal  USD
        2022-01-01 open Assets:Federal:401k     IRAUSD
        2022-01-01 open Expenses:Taxes:401k     IRAUSD
        2022-01-01 open Assets:Vacation         VACHR
        2022-01-01 open Income:Vacation         VACHR
        2022-01-01 open Expenses:Vacation       VACHR
        2022-01-01 open Expenses:Tests          USD

        2022-01-01 * "ACME" "Salary"
          Assets:Checking           10.00 USD
          Income:ACME              -11.00 USD
          Expenses:Taxes:Federal     1.00 USD
          Assets:Federal:401k       -2.00 IRAUSD
          Expenses:Taxes:401k        2.00 IRAUSD
          Assets:Vacation               5 VACHR
          Income:Vacation              -5 VACHR

        2022-01-02 * "Holidays"
          Assets:Vacation              -1 VACHR
          Expenses:Vacation

        2022-01-03 * "Test"
          Assets:Checking            3.00 USD
          Expenses:Tests

        """

        sql_query = r"""
          SELECT
            account,
            sum(position) AS amount
          WHERE root(account, 1) = '{0}'
          GROUP BY 1
          ORDER BY 2 DESC
        """

        rtypes, rrows = query.run_query(entries, options, sql_query, 'Expenses', numberify=True)
        columns = [c.name for c in rtypes]
        self.assertEqual(columns, ['account', 'amount (USD)', 'amount (VACHR)', 'amount (IRAUSD)'])
        self.assertEqual(len(rrows[0]), 4)


if __name__ == '__main__':
    unittest.main()
